private import std.string;
private import hell2;
private import sysinfo;
private import token;
private import particle;
private import fmfloader;
private import natsuki2;

/**
 * 敵
 */
class Enemy : Token
{
public:
	enum {
		STATE_APPEAR,  // 出現
		STATE_NORMAL,  // 通常
		STATE_DAMAGE,  // ダメージ
		STATE_SUICIDE, // 自爆演出
		STATE_WARP,    // ワープ
	};
	const int   TIMER_REFLECT = 30;   // 打ち返しダメージのタイマ
	const int   TIMER_DAMAGE  = 3;    // ボスのダメージ点滅タイマ
	const int   TIMER_WARP    = 120;  // ワープタイマ
	const int   ID_MAX        = 32;
	static TokenManager!(Enemy) parent; // 親
	static Player               target; // プレイヤー
	static int[]                HP_TBL;
	static int TimerAppear = 120; // 敵出現タイマ
	static int Rank;              // ランク
	// 変数
	int    id;
	int    timer;
	int    state;
	float  decay;         // 移動の減衰量
	int    ratio;         // スコア倍率
	int    hpNext;        // 体力(じわじわ減る用)
	int    hp;            // 体力
	int    hpmax;         // 体力最大値
	int    tAppearMax;    // 出現タイマ開始
	int    tReflect;      // 打ち返し弾ダメージタイマ
	int    reflectDamage; // 打ち返し弾のダメージ量
	int    attr;          // 付加属性
	int    tPast;         // 生存時間
	float  direction;     // 移動方向２
	float  speed;         // 速さ
	int    param1;        // 汎用パラメータ１
	int    tDamage;       // ダメージ点滅用タイマ
	int    tWarp;         // ワープタイマ
	float  prevX;         // ワープ元(X)
	float  prevY;         // ワープ元(Y)
	float  nextX;         // ワープ先(X)
	float  nextY;         // ワープ先(Y)
	int    chain;         // 連鎖回数
private:
	int    _score;
public:
	this()
	{
		HP_TBL = new int[ID_MAX];
	}
	void create(int id, float x, float y, float deg=0, float speed=0)
	{
		this.id    = id;
		this.speed = speed;
		int w   = GH.ENEMY_SIZE[id];
		int h   = GH.ENEMY_SIZE[id];
		_score  = GH.ENEMY_SCORE[id];
		decay   = GH.ENEMY_DECAY[id];
		hp      = HP_TBL[id];
		hpNext  = hp;
		hpmax   = hp;
		super.create(x, y, w, h, w, h, Hell_cosEx(deg)*speed, -Hell_sinEx(deg)*speed);
		timer         = 0;
		state         = STATE_NORMAL;
		ratio          = 1;
		rot           = 0;
		attr          = 0;
		tPast         = 0;
		tReflect      = 0;
		reflectDamage = 0;
		param1        = 0;
		tDamage       = 0;
		tWarp         = 0;
		chain         = 0;
	}
	// 敵HPの初期値を設定する
	static void setHp(int id, int hp)
	{
		HP_TBL[id] = hp;
	}
	// 敵出現タイマーの設定
	static void setAppearTimer(int t)
	{
		TimerAppear = t;
	}
	static void vanishAll()
	{
		foreach(e; parent.getArray())
		{
			if(!e.exist)   continue;
			if(e.isBoss()) continue;
			e.ratio = SceneMain.Ratio;
			int score = e.vanish();
			SceneMain.Score += score;
		}
	}
	static Enemy add(int id, float x, float y, float deg=0, float speed=0)
	{
		Enemy enemy = parent.getToken();
		if(!enemy) return null;
		enemy.create(id, x, y, deg, speed);
		return enemy;
	}
	// 出現タイマーあり
	static Enemy addEx(int id, float x, float y, float deg=0, float speed=0)
	{
		Enemy enemy = parent.getToken();
		if(!enemy) return null;
		enemy.create(id, x, y, deg, speed);
		enemy.beginAppear(TimerAppear);
		return enemy;
	}
	/**
	 * FMFファイルから敵を生成
	 */
	static void loadFMF(char[] filepath)
	{
		scope fmf = new FMFLoader(filepath);
		Layer2D layer1 = fmf.getLayer(0); // 敵レイヤーは「0」
		Layer2D layer2 = fmf.getLayer(1); // 敵属性レイヤーは「1」
		for(int j = 0; j < layer1.height; j++)
		{
			for(int i = 0; i < layer1.width; i++)
			{
				int id = layer1.get(i, j);
				if(id == 0) continue;
				Enemy e = add(id, i*fmf.chipwidth+fmf.chipwidth/2, j*fmf.chipheight+fmf.chipheight/2);
				if(e is null) continue;
				e.beginAppear(TimerAppear);
				int attr = layer2.get(i, j);
				if(attr == 0) continue;
				e.setAttribute(attr);
			}
		}
	}
	// ザコの生存数をカウントする
	static int countZaco()
	{
		int ret = 0;
		foreach(e; parent.getArray())
		{
			if(!e.exist)   continue;
			if(e.isBoss()) continue;
			ret++;
		}
		return ret;
	}
	// 指定IDの敵をカウントする
	static int countId(int id)
	{
		int ret = 0;
		foreach(e; parent.getArray())
		{
			if(!e.exist)   continue;
			if(e.id != id) continue;
			ret++;
		}
		return ret;
	}
	static void updateBoss()
	{
		foreach(e; parent.getArray())
		{
			if(!e.exist)    continue;
			if(!e.isBoss()) continue;
			e.update();
		}
	}
	static Enemy getBoss()
	{
		foreach(e; parent.getArray())
		{
			if(!e.exist)    continue;
			if(!e.isBoss()) continue;
			return e;
		}
		return null;
	}
	static bool isBossSuicide()
	{
		foreach(e; parent.getArray())
		{
			if(!e.exist)    continue;
			if(!e.isBoss()) continue;
			return e.isSuicide();
		}
		return false;
	}
	static int randint(int a, int b) { return Hell_randint(a, b); }
	// 打ち返し弾の数を取得
	int getReflectCount()
	{
		int cnt = GH.REFLECT_BULLET_CNT * ratio;
		if(cnt < 1) cnt = 1;
		return cnt;
	}
	float getAim()
	{
		float dx = target.x - x;
		float dy = target.y - y;
		return Hell_atan2Ex(-dy, dx);
	}
	// 出現演出モードへ
	void beginAppear(int t)
	{
		state      = STATE_APPEAR;
		timer      = t;
		tAppearMax = t;
	}
	void setAttribute(int attr)
	{
		this.attr = attr;
	}
	void beginReflect(int val)
	{
		tReflect      =  TIMER_REFLECT;
		reflectDamage += val;
	}
	// ショットが当たるかどうか
	bool isHitShot()
	{
		return state == STATE_NORMAL;
	}
	// 打ち返しが当たるかどうか
	bool isHitReflect()
	{
		switch(state)
		{
		case STATE_NORMAL:
		case STATE_DAMAGE:
			return true;
		default:
			return false;
		}
	}
	bool isDamage()  { return state == STATE_DAMAGE;  } // ダメージ中かどうか
	bool isSuicide() { return state == STATE_SUICIDE; } // 自爆中かどうか
	// バリアがあるかどうか
	bool isBarrier()
	{
		if(!isBoss())              return false; // IDがザコ
		if(100*hpNext/hpmax <= 30) return false; // HPが30%以下
		return true;
	}
	// 倒せそうかどうか
	bool isDanger()
	{
		return (100*hpNext/hpmax <= 10); // HPが10%以下
	}
	bool isDead() { return hpNext < 1;   }
	bool isBoss() { return attr == 2; } // ボスかどうか
	void update()
	{
		if(tDamage > 0) tDamage--;
		if(tReflect > 0)
		{
			tReflect--;
			if(isBoss() && tReflect == 0)
			{
				Particle.addMessage2(x, y, format("%d", reflectDamage), [0xff, 0x00, 0x00], reflectDamage);
				reflectDamage = 0;
			}
		}
		if(tDamage == 0)
		{
			if(hpNext < hp)
			{
				hp -= 3;
				if(hp < hpNext) hp = hpNext;
			}
		}
		tPast++;
		switch(state)
		{
		case STATE_APPEAR:
			doAppear();
			break;
		case STATE_NORMAL:
			doNormal();
			move();
			break;
		case STATE_DAMAGE:
			doDamage();
			move();
			break;
		case STATE_SUICIDE:
			doSuicide();
			move();
			break;
		case STATE_WARP:
			doWarp();
			move();
			break;
		}
	}
	void move(float deg, float speed)
	{
		vx = Hell_cosEx(deg)*speed;
		vy = -Hell_sinEx(deg)*speed;
	}
	// 直角に動く
	void moveOrtho(float speed)
	{
		move(getAim()+(randint(0, 1) == 0 ? 90 : -90), speed);
	}
	// ターゲットの方を向く
	void rotTarget(float rotSpeed)
	{
		float deg = getAim();
		float r   = deg - rot;
		if(Hell_abs(r) < 3) { rot = deg; }
		else
		{
			if(Hell_abs(r) > 180)
			{
				if(rot > deg) { rot -= 360; }
				else          { rot += 360; }
			}
			rot = rot*(1 - rotSpeed) + deg*rotSpeed;
		}
	}
	void doAppear()
	{
		timer--;
		if(timer < 1) state = STATE_NORMAL;
	}
	void doNormal()
	{
		switch(id)
		{
		case 1:
			// なす
			updateNasu();
			break;
		case 2:
			// 5箱
			updateGo();
			break;
		case 3:
			// だいこん
			updateDaikon();
			break;
		case 4:
			// たこやき
			updateTako();
			break;
		case 5:
			// いちご
			updateIchigo();
			break;
		case 6:
			// ポッキー
			updatePokky();
			break;
		case 15:
			// 巨大なす
			updateBigNasu();
			break;
		case 16:
			// プリン
			vx = vx * 0.95;
			vy = vy * 0.95;
			timer++;
			if(timer%130 == 0) Bullet.addNWay(x, y, 1.2f, 5, -45, 45);
			if(timer%200 == 0) moveOrtho(8);
			if(timer%60 == 0)
			{
				if(countZaco() < 3)
				{
					Enemy.add(1, x, y, randint(0,   179), 10);
					Enemy.add(1, x, y, randint(180, 359), 10);
				}
			}
			break;
		case 17:
			// 牛乳
			updateMilk();
			break;
		case 18:
			// X箱
			updateXBox();
			break;
		}
		if(isBoss())
		{
			switch(Rank)
			{
			case 0:
				if(countZaco() == 0)
				{
					// プチ発狂
					Enemy.add(1, x, y, getAim()+90, 10);
					Enemy.add(1, x, y, getAim()-90, 10);
					if(isBarrier())
					{
						if(Hell_randint(0, 1) == 0) Bullet.addAllRange(x, y, 1.2f, 32); // 全方位弾
						else                        Bullet.addNWay(x, y, 1.5f, 7, -60, 60);

					}
					else
					{
						Bullet.addAllRange(x, y, 1.8f, 8); // 8方位弾
						Bullet.addAllRange(x, y, 1.2f, 8); // 8方位弾
						Bullet.addAllRange(x, y,  .6f, 8); // 8方位弾
					}
				}
				break;
			case 1:
				if(timer%200 == 0) callZaco(2, 1); // 5箱は常に１つ生成
				if(timer%160 == 0) callZaco(4, 3); // たこやきは常に３つ生成
				if(!isBarrier())
				{
					// 発狂
					callZaco(2, 4);
					callZaco(4, 5);
				}
				break;
			case 2:
				if(timer%240 == 0) callZaco(2, 1); // 5箱は常に１つ生成
				if(timer%200 == 0) callZaco(5, 1); // いちごを常に１つ生成
				if(timer%160 == 0) callZaco(4, 3); // たこやきは常に２つ生成
				if(!isBarrier())
				{
					// 発狂
					callZaco(4, 4);
					callZaco(5, 3);
				}
				break;
			}
		}
	}
	// 指定IDの敵が一定数以下なら生成
	void callZaco(int id, int count)
	{
		int cnt = countId(id);
		for(int i = cnt; i < count; i++)
		{
			Enemy.addEx(id, randint(32, GH.WIDTH-32), randint(32, GH.HEIGHT-32));
		}
	}
	// なすの更新
	void updateNasu()
	{
		timer++;
		if(Rank == 0) { vx = vx * 0.95; vy = vy * 0.95; return; }
		
		if(timer == 20)
		{
			float dx    = GH.WIDTH/2  - x;
			float dy    = GH.HEIGHT/2 - y;
			int   speed = 50 + 20*(Rank-1);
			move(Hell_atan2Ex(-dy, dx), randint(speed, speed+10)/100f);
		}
	}
	// 5箱の更新
	void updateGo()
	{
		vx = vx * 0.95;
		vy = vy * 0.95;
		timer++;
		if(timer > 540)
		{
			// 回転減速
			int t = 180-(timer%180);
			rot = t*t/10;
			timer %= 720;
		}
		else if(timer > 360)
		{
			// 高速回転
			timer++;
			int t = timer%180;
			rot = t*t/10;
			if(timer == 540) Bullet.addAllRange(x, y);
		}
		else
		{
			if(timer%120 == 0) Bullet.addNWay(x, y, 1f);
		}
	}
	// だいこんの更新
	void updateDaikon()
	{
		rot     = Hell_atan2Ex(-vy, vx);
		float d = speed - Hell_distance(vx, vy);
		if(d > 0.2f)
		{
			// 基準のスピードより遅いので補正
			vx += 0.1f * d * Hell_cosEx(rot);
			vy += 0.1f * d * -Hell_sinEx(rot);
		}
	}
	// たこやきの更新
	void updateTako()
	{
		vx = vx * 0.95;
		vy = vy * 0.95;
		timer++;
		if(timer <= 0) return;
		
		switch(Rank)
		{
		case 1:
			rot = timer*timer/10;
			switch(timer%180)
			{
			case 160:
				// 移動方向を覚えておく
				direction = getAim();
				break;
			case 0:
				// 移動する
				move(direction, 8);
				timer = -randint(100, 120);
				break;
			default:
				break;
			}
			break;
		default:
			rot = timer*timer/5;
			switch(timer%90)
			{
			case 80:
				// 移動方向を覚えておく
				direction = getAim();
				break;
			case 0:
				// 移動する
				rot = 0;
				move(direction, 7);
				timer = -randint(50, 60);
				break;
			default:
				break;
			}
			break;
		}
	}
	// いちごの更新
	void updateIchigo()
	{
		timer++;
		float d = Hell_distance(vx, vy);
		if(timer < 100)
		{
			rot += 18;
			vx *= 0.95f;
			vy *= 0.95f;
			if(timer == 99)
			{
				rot = getAim();
			}
		}
		else if(timer < 340)
		{
			vx = d * Hell_cosEx(rot);
			vy = d * -Hell_sinEx(rot);
			rotTarget(randint(8, 12)/1000f);
			if(d < 2f)
			{
				// 基準のスピードより遅いので補正
				vx += 0.1f * Hell_cosEx(rot);
				vy += 0.1f * -Hell_sinEx(rot);
			}
		}
		else if(timer < 580)
		{
			rot = Hell_atan2Ex(-vy, vx);
			if(timer == 579) timer = 0;
		}
	}
	// ポッキーの更新
	void updatePokky()
	{
		if(timer < 270)
		{
			vx *= 0.97f;
			vy *= 0.97f;
			if(Rank == 2)
			{
				rot += (135 - timer%135)/6;
				if(timer%135 == 0) direction = getAim();
				if(timer%135 < 30)
				{
					if(timer%6 == 0) Bullet.add(x, y, direction, 2f);
				}
			}
			else
			{
				rot += (90 - timer%90)/4;
				if(timer%90 == 0) direction = getAim();
				if(timer%90 < 30)
				{
					if(timer%6 == 0) Bullet.add(x, y, direction, 2f);
				}
			}
		}
		else if(timer == 270)
		{
			rot += (90 - timer%90)/4;
			for(int i = 0; i < 8; i++)
			{
				float dx = 40*Hell_cosEx(i*45);
				float dy = 40*-Hell_sinEx(i*45);
				Bullet.addAim(x+dx, y+dy, 3f);
			}
			if(param1 >= 1)
			{
				// 3回目で自爆
				vanish();
			}
		}
		else if(timer < 360)
		{
			rot += (90 - timer%90)/4;
		}
		else
		{
			timer = -1;
			param1++;
		}
		timer++;
	}
	// 牛乳の更新
	void updateMilk()
	{
		vx = vx * 0.95;
		vy = vy * 0.95;
		timer++;
		switch(timer)
		{
		case 200:
		case 440:
			moveOrtho(8);
			break;
		case 360:
			float r = getAim() - 30;
			if(isBarrier())
			{
				for(int i = 0; i < 3; i++)
				{
					float r2 = r + i * 30;
					Enemy.add(3, x, y, r2, 3); // だいこんを飛ばす
					for(int j = 0; j < 8; j++)
					{
						Bullet.add(x, y, r2, 1.4f + j*0.20f);
					}
				}
			}
			else
			{
				// 発狂
				for(int i = 0; i < 8; i++)
				{
					float r2 = r + i * 45 + 22.5f;
					Enemy.add(3, x, y, r2, 3); // だいこんを飛ばす
					for(int j = 0; j < 8; j++)
					{
						Bullet.add(x, y, r2, 1.4f + j*0.20f);
					}
				}
			}
			break;
		case 520:
			Enemy.add(6, x, y, getAim() - 90, 8);
			break;
		case 540:
			Enemy.add(6, x, y, getAim() - 180, 8);
			break;
		case 560:
			Enemy.add(6, x, y, getAim() + 90, 8);
			timer = 0;
			break;
		default:
			break;
		}
	}
	// X箱の更新
	void updateXBox()
	{
		vx = vx * 0.95;
		vy = vy * 0.95;
		timer++;
		if(timer%200 == 0) moveOrtho(8);
		if(timer%360 == 0)
		{
			float r = getAim();
			Enemy.add(3, x, y, r, 3); // だいこんを飛ばす
			for(int i = 1; i < 8; i++)
			{
				Bullet.add(x, y, r+i*10, (8-i)/3f);
				Bullet.add(x, y, r-i*10, (8-i)/3f);
			}
		}
	}
	// 巨大なす
	void updateBigNasu()
	{
		vx = vx * 0.95;
		vy = vy * 0.95;
		timer++;
		if(timer < 40)
		{
			if(timer == 0) moveOrtho(8);
			int t = timer;
			// ウィップ大根
			if(t%5 == 0)
			{
				Enemy.add(3, x, y, getAim(), .7f + .4f*t/5);
			}
		}
		else if(timer < 360)
		{
			if(timer%200 == 0) moveOrtho(8);
		}
		else if(timer < 400)
		{
			if(timer == 360) moveOrtho(8);
			int t = timer - 360;
			// ウィップ大根
			if(t%5 == 0)
			{
				Enemy.add(3, x, y, getAim(), .7f + .4f*t/5);
			}
		}
		else if(timer < 760)
		{
			if(timer%200 == 0) moveOrtho(8);
		}
		else if(timer < 1120)
		{
			if(timer%4 != 0) return;
			int r = timer%360;
			for(int i = 0; i < 4; i++)
			{
				Bullet.add(x, y, r+i*90, 2f);
			}
		}
		else if(timer < 1480)
		{
			if(timer%200 == 0) moveOrtho(8);
		}
		else if(timer < 1540)
		{
			switch(timer)
			{
			case 1480:
				Enemy.add(6, x, y, getAim() - 90, 8);
				break;
			case 1500:
				Enemy.add(6, x, y, getAim() - 180, 8);
				break;
			case 1520:
				Enemy.add(6, x, y, getAim() + 90, 8);
				break;
			default:
				break;
			}
		}
		else
		{
			timer = 0;
		}
	}

	void doDamage()
	{
		vx = vx * decay;
		vy = vy * decay;
		float d = Hell_distance(vx, vy);
		rot += d*4;
		timer++;
		if(d < 0.5)
		{
			rot   = 0;
			ratio  = 1;
			vx    = 0;
			vy    = 0;
			state = STATE_NORMAL;
			timer = 0;
			switch(id)
			{
			case 3:
				vanish();
				break;
			default:
				break;
			}
		}
	}
	void doSuicide()
	{
		switch(id)
		{
		case 15:
			vx *= 0.97f;
			vy *= 0.97f;
			rot += 10;
			int r = cast(int)rot;
			for(int i = 0; i < 8; i++)
			{
				int   w     = 48;
				float speed = 32f;
				Particle particle = Particle.add("destroy3", x, y, w, w, r, speed, 60, [0xff, 0xff, 0xff]);
				if(particle)
				{
					particle.rot   = r;
					particle.speed = speed;
				}
				r += 45;
			}
			if(timer%3 == 0)
			{
				float rot = randint(0, 359);
				Hell_stopWAV(6);
				Hell_playWAV("destroy1", 0, 6);
			}
			break;
		case 16:
			x += randint(-timer, timer);
			y += randint(-timer, timer);
			Particle.addDestroy(x, y, [0xff, 0x40, 0x40]);
			if(timer%3 == 0)
			{
				Hell_stopWAV(6);
				Hell_playWAV("destroy1", 0, 6);
			}
			break;
		case 17:
			rot += 20;
			vx += 4 * Hell_cosEx(rot);
			vy += 4 * -Hell_sinEx(rot);
			if(timer%3 == 0)
			{
				Particle.addCut(x, y, randint(0, 359), randint(10, 20), [0xff, 0x40, 0x40]);
				Hell_stopWAV(6);
				Hell_playWAV("destroy1", 0, 6);
			}
			break;
		case 18:
			rot += 20;
			Particle.addDestroy(x, y, [0xff, 0x40, 0x40]);
			if(timer%3 == 0)
			{
				Hell_stopWAV(6);
				Hell_playWAV("destroy1", 0, 6);
			}
			break;
		}
		timer--;
		if(timer < 1)
		{
			SceneMain.Score += vanish();
		}
	}
	// ワープ
	void doWarp()
	{
		x = prevX + (nextX - prevX) * (1 - (1f * tWarp / TIMER_WARP));
		y = prevY + (nextY - prevY) * (1 - (1f * tWarp / TIMER_WARP));
		tWarp = tWarp * 90 / 100;
		if(tWarp < 1) state = STATE_NORMAL;
	}

	// 吹き飛ばす
	void blow(float deg, float speed)
	{
		vx = Hell_cosEx(deg)  * speed;
		vy = -Hell_sinEx(deg) * speed;
	}
	void _damage(int val)
	{
		hpNext -= val;
	}
	void damage(float deg, float speed, int val)
	{
		switch(state)
		{
		case STATE_APPEAR:
			break;
		case STATE_NORMAL:
			blow(deg, speed);
			if(!isBoss())
			{
				state = STATE_DAMAGE;
				break;
			}
			
			// ボス
			tDamage = TIMER_DAMAGE;
			if(id == 15)
			{
				if(val > 100) val = 100;
			}
			if(isBarrier())
			{
				// バリアあり
				Particle.addMessage2(x, y, format("%d", val), [0xff, 0x00, 0x00], val);
				_damage(val);
				if(hpNext < 1) hpNext = 1;
				if(!isBarrier())
				{
					Hell_playWAV("crash");
					Particle.addBarrier(x, y);
				}
				break;
			}
			Particle.addMessage2(x, y, format("%d", val), [0xff, 0x00, 0x00], val);
			_damage(val);
			if(hpNext < 1)
			{
				// 自爆開始
				beginSuicide();
			}
			break;
		case STATE_DAMAGE:
			if(GH.SHOT_DAMAGE_ENEMY)
			{
				// ダメージ中もヒットあり
				blow(deg, speed);
				state = STATE_DAMAGE;
			}
			break;
		default:
			break;
		}
	}
	
	// ダイレクトダメージ(打ち返しヒット)
	void damageEx(float deg, float speed, int val)
	{
		switch(state)
		{
		case STATE_NORMAL:
		case STATE_DAMAGE:
			beginReflect(val);
			// 移動量に足しこむ
			vx += Hell_cosEx(deg)*speed;
			vy += -Hell_sinEx(deg)*speed;
			bool bBarrier = isBarrier();
			_damage(val);
			if(hpNext < 1)
			{
				hpNext = 1;
				if(isBoss()) beginSuicide();
				else         hpNext = 0;
			}
			if(bBarrier && !isBarrier())
			{
				Hell_playWAV("crash");
				Particle.addBarrier(x, y);
			}
			tDamage = TIMER_DAMAGE;
			break;
		default:
			break;
		}
		if(id == 15)
		{
			// だいこんガード
			float rot = randint(0, 359);
			for(int i = 0; i < 16; i++)
			{
				Enemy.add(3, x, y, rot, .7f + .4f*i);
				rot += 22.5f;
			}
			if(!isBarrier())       return;
			if(hp - hpNext < 1000) return;
			// ワープ開始
			prevX = x;
			prevY = y;
			int w2 = GH.WIDTH/2;
			int h2 = GH.HEIGHT/2;
			if(x <  w2 && y <  h2) { nextX = randint(w2, w2*2-cast(int)w); nextY = randint(h2, h2*2-cast(int)w); }
			if(x >= w2 && y <  h2) { nextX = randint(cast(int)w,  w2);     nextY = randint(h2, h2*2-cast(int)w); }
			if(x <  w2 && y >= h2) { nextX = randint(w2, w2*2-cast(int)w); nextY = randint(cast(int)w,  h2); }
			if(x >= w2 && y >= h2) { nextX = randint(cast(int)w,  w2);     nextY = randint(cast(int)w,  h2); }
			tWarp = TIMER_WARP;
			state = STATE_WARP;
		}
	}
	void beginSuicide()
	{
		switch(id)
		{
		case 15:
			float dx  = GH.WIDTH/2  - x;
			float dy  = GH.HEIGHT/2 - y;
			float rot = Hell_atan2Ex(-dy, dx);
			vx = 10 * Hell_cosEx(rot);
			vy = 10 * -Hell_sinEx(rot);
			break;
		case 16:
		case 17:
			vx = 0;
			vy = 0;
			break;
		case 18:
			float deg = Hell_atan2Ex(-vy, vx);
			move(deg, 20);
			break;
		}
		state = STATE_SUICIDE;
		timer = GH.MAIN_SUICIDE_TIMER;
	}
	void move()
	{
		if(x < w)             { x = w;          vx = -vx; }
		if(y < h)             { y = h;          vy = -vy; }
		if(x > GH.WIDTH  - w) { x = GH.WIDTH  - w; vx = -vx; }
		if(y > GH.HEIGHT - h) { y = GH.HEIGHT - h; vy = -vy; }
		super.move();
	}
	int vanish()
	{
		// アイテムドロップ
		switch(attr)
		{
		case 1:
			Item.addBananaEx(x, y, 1);
			break;
		default:
			break;
		}
		Particle.addDestroy(x, y);
		int score = _score;
		{
			float dx    = GH.WIDTH/2  - x;
			float dy    = GH.HEIGHT/2 - y;
			float deg   = Hell_atan2Ex(-dy, dx);
			Particle.addNumber(x, y, deg, 5, score, ratio);
		}
		exist = false;
		if(isBoss())
		{
			Hell_stopWAV(7);
			Hell_playWAV("destroy3", 0, 7);
			if(tReflect > 0)
			{
				// ダメージ量表示
				tReflect = 0;
				Particle.addMessage2(x, y, format("%d", reflectDamage), [0xff, 0x00, 0x00], reflectDamage);
				reflectDamage = 0;
			}
			Particle.addDestroy2(x, y);
			for(int i = 0; i < 2; i++)
			{
				Particle.addCut(x, y, i*90, 0);
			}
			vanishAll();
			// 弾消し
			Bullet.vanishAll();
		}
		return score;
	}
	static void drawBoss()
	{
		foreach(e; parent.getArray())
		{
			if(!e.exist)    continue;
			if(!e.isBoss()) continue;
			e.draw();
		}
	}
	void draw()
	{
		float sx = 1.0f;
		float sy = 1.0f;
		int r = 0xff;
		int g = 0xff;
		int b = 0xff;
		switch(state)
		{
		case STATE_APPEAR:
			float ratio = cast(float)(tAppearMax - timer)/tAppearMax;
			if(ratio > 0.8f && timer%8 < 4) return;
			r = g = b = cast(int)(0xff * ratio);
			Hell_setAlpha(HELL_ALPHA_ADD);
			sx = 0.5f + 0.5f * ratio;
			sy = 0.5f + 0.5f * ratio;
			break;
		case STATE_NORMAL:
			if(tDamage > 0) Hell_setAlpha(HELL_ALPHA_ADD); // ダメージ点滅
			else            Hell_setAlpha(HELL_ALPHA_NORMAL);
			if(timer%32 < 20)
			{
				g = isDanger() ? 0 : 0xff; // 倒せそうで赤くする
				b = isDanger() ? 0 : 0xff; // 倒せそうで赤くする
			}
			break;
		case STATE_DAMAGE:
			Hell_setAlpha(HELL_ALPHA_NORMAL);
			//if(timer%16 < 8)
			{
				//Hell_setAlpha(HELL_ALPHA_ADD);
				//Hell_drawCircle(x, y, w, 0, 0x80, 0xff, 0x80);
				r = 0x00;
				b = 0x00;
			}
			sx = 1f + 0.5f * Hell_sinEx((timer%30)*6);
			break;
		case STATE_SUICIDE:
			Hell_setAlpha(HELL_ALPHA_XOR);
			break;
		case STATE_WARP:
			Hell_setAlpha(HELL_ALPHA_ADD);
			break;
		}
		// キャラの描画
		int ox = 0;
		int oy = 0;
		int ow = 0;
		int oh = 0;
		char[] imgName; // 画像の名前
		if(id < 16) { imgName = "enemy";  }
		else        { imgName = "enemy2"; }
		switch(id)
		{
		case 1:  ox = 0;   oy = 0;   ow = 128; oh = 128; break; // なす
		case 2:  ox = 128; oy = 0;   ow = 128; oh = 128; break; // 5
		case 3:  ox = 0;   oy = 128; ow = 128; oh = 128; break; // だいこん
		case 4:  ox = 128; oy = 128; ow = 128; oh = 128; break; // たこやき
		case 5:  ox = 386; oy = 0;   ow = 128; oh = 128; break; // いちご
		case 6:  ox = 256; oy = 128; ow = 128; oh = 128; break; // ポッキー
		case 15: ox = 0;   oy = 0;   ow = 128; oh = 128; break; // 巨大なす
		case 16: ox = 0;   oy = 0;   ow = 256; oh = 256; break; // プリン
		case 17: ox = 256; oy = 0;   ow = 256; oh = 256; break; // 牛乳
		case 18: ox = 0;   oy = 256; ow = 256; oh = 256; break; // X箱
		}
		switch(id)
		{
		case 1:
		case 15:
		case 16:
		case 17:
		case 18:
			// 左右に振り向く
			int rev = ((isSuicide() || isDamage() || timer%120 < 60) ? 1 : -1);
			if(state == STATE_WARP)
			{
				Hell_drawTextureEx(imgName, prv2X, prv2Y, ox, oy, ow, oh, sx*rev*w*2/ow, sy*h*2/oh, rot, 0, 0, b, 64);
				Hell_drawTextureEx(imgName, prvX,  prvY,  ox, oy, ow, oh, sx*rev*w*2/ow, sy*h*2/oh, rot, 0, 0, b, 128);
				Hell_drawTextureEx(imgName, x,     y,     ox, oy, ow, oh, sx*rev*w*2/ow, sy*h*2/oh, rot, 0, 0, b);
			}
			else
			{
				Hell_drawTextureEx(imgName, x, y, ox, oy, ow, oh, sx*rev*w*2/ow, sy*h*2/oh, rot, r, g, b);
			}
			// drawSub(imgName, ox, oy, ow, oh, r, g, b, rev);
			break;
		default:
			Hell_drawTextureEx(imgName, x, y, ox, oy, ow, oh, sx*w*2/ow, sy*h*2/oh, rot, r, g, b);
			break;
		}
		
		drawAttribute();
		drawRatio();
		if(state != STATE_WARP)
		{
			if(isBarrier()) { drawBarrier(); } // バリアの描画
		}
		if(isBoss() && !isSuicide()) { drawHp();      }
	}
	void drawRatio()
	{
		if(isBoss())  return;
		if(ratio < 2) return;
		int idx = 0;
		switch(ratio)
		{
		case 2:  idx = 1; break;
		case 4:  idx = 2; break;
		case 8:  idx = 3; break;
		case 16: idx = 4; break;
		case 32: idx = 5; break;
		default: return;
		}
		Hell_setAlpha(HELL_ALPHA_NORMAL);
		int oy = 16*idx;
		Hell_drawTextureEx("font2", x, y, 0, oy, 36, 16);
	}
	// バリアの描画
	void drawBarrier()
	{
		Hell_setAlpha(HELL_ALPHA_ADD);
		int a = 0xff*hp/hpmax - 0xff*10/100;
		if(state == STATE_APPEAR)
		{
			for(int i = 0; i < 7; i++)
			{
				int t     = timer + i*i;
				int width = cast(int)((w*4)*t*t/(TimerAppear*TimerAppear));
				Hell_drawCircle(x, y, width, 6, a, 0, a, randint(0x80, 0xff));
				a -= 10;
				if(a < 0) a = 0;
			}
			if(timer < 20)
			{
				a = 0xff*(20-timer)/20;
				Hell_drawCircle(x, y, w*1.2f, 0, a, 0, a, randint(0x80, 0xff));
			}
		}
		else
		{
			Hell_drawCircle(x, y, w*1.2f, 0, a, 0, a, randint(0x80, 0xff));
		}
	}
	// 属性の描画
	void drawAttribute()
	{
		float dx, dy;
		switch(attr)
		{
		case 1:
			// バナナを持っているよ
			//Hell_setAlpha(HELL_ALPHA_NORMAL);
			for(int i = 0; i < 2; i++)
			{
				dx = w * Hell_cosEx(tPast*2 + i*180);
				dy = w * -Hell_sinEx(tPast*2 + i*180);
				Hell_drawTextureEx("enemy", x+dx, y+dy, 256, 64, 64, 64, .3f, .3f, 0);
			}
			break;
		default:
			break;
		}
	}
	void drawHp()
	{
		if(GH.ENEMY_HP_STRING)
		{
			// 文字で
			Hell_drawFontEx(cast(int)x, cast(int)y, .5f, 0xff, 0xff, 0xff, 0xff, "HP:%d", hp);
		}
		else
		{
			Hell_setAlpha(HELL_ALPHA_NORMAL);
			int W = cast(int)w;
			int H = 6;
			float px = x - W/2;
			float py = y + w*0.7f;
			int   w2 = W*hp/hpmax;
			if(w2 < 1) w2 = 1;
			int   w3 = W*hpNext/hpmax;
			if(w3 < 1) w3 = 1;
			int a = 0xa0;
			Hell_drawRect(cast(int)px-2, cast(int)py-2, W+4, H+4, 0, 0, 0,    0,    0, a);
			Hell_setAlpha(HELL_ALPHA_NORMAL);
			Hell_drawRect(cast(int)px,   cast(int)py,   w2,  H,   0, 0, 0xff, 0,    0);
			Hell_drawRect(cast(int)px,   cast(int)py,   w3,  H,   0, 0, 0xff, 0xff, 0);
		}
	}
}
