private import std.string;
private import hell2;
private import sysinfo;
private import token;
private import enemy;
private import particle;
private import levelmanager;
private import configparser;
private import message;

private int    g_fps = 1000/60; // fps

/**
 * 照準
 */
class Aim : Token
{
public:
	static Player parent; // 親
	int           timer;  // アニメーションタイマ
	int           tShot;  // ショット制限タイマ
	// 生成
	void create(Player player, float x, float y)
	{
		int SIZE = GH.AIM_SIZE;
		parent = player;
		timer  = 0;
		tShot  = 0;
		super.create(x, y, SIZE, SIZE, SIZE, SIZE);
	}
	// ショットを撃てるかどうか
	bool isShot()
	{
		if(tShot > 0)         return false;
		if(parent.isDamage()) return false; // ダメージ中は撃てない
		return true;
	}
	// ショット発射
	void shot()
	{
		float dx = x - parent.x;
		float dy = y - parent.y;
		//Shot.addNormal(x, y, Hell_atan2Ex(-dy, dx), 0);
		int BASE = GH.SHOT_RELOAD_TIMER;
		Shot.addNormal2(x, y, 1f*(BASE-tShot)/BASE);
		Hell_stopWAV(4);
		Hell_playWAV("shot", 0, 4);
		tShot = GH.SHOT_RELOAD_TIMER;
	}
	// 更新
	void update()
	{
		updateAnime();
		if(tShot > 0)
		{
			tShot--;
			if(tShot <= 0)
			{
				tShot = 0;
				Particle.addShot(x, y); // ショット発射可能に
			}
		}
		move();
	}
	// アニメーション更新
	void updateAnime()
	{
		rot += 2f;
		if(rot > 360f) rot -= 360f;
		timer++;
		x = Hell_getMouseX();
		y = Hell_getMouseY();
	}
	// 描画
	void draw(bool bDanger=false)
	{
		Hell_setAlpha(HELL_ALPHA_ADD);
		int r = 0x00;
		int g = 0xff;
		int b = 0x00;
		if(bDanger)
		{
			r = 0xff * (timer%32)/32;
			b = 0xff * (timer%32)/32;
			Hell_drawCircle(x, y, Hell_randint(1, GH.AIM_SIZE), 3, 0, 0xff, 0);
		}
		int width = isShot() ? 0 : 1;
		// 十字の描画
		drawCross(width);
		
		float d = 7*GH.AIM_SIZE/8; // 円の半径
		if(isShot()) { d = (GH.AIM_SIZE - 2) + 4f*Hell_sinEx((timer*2)%360); } // 円の伸び縮み
		float rot = isShot() ? rot : 0; // 回転角度
		// 円の描画
		drawCircle(width, d, rot, r, g, b);
	}
	// 十字の描画
	void drawCross(int width)
	{
		int s2 = 3*GH.AIM_SIZE/5;
		Hell_drawRectEx(cast(int)x, cast(int)y, s2, 4, width, -rot/2,     0, 0xff, 0);
		Hell_drawRectEx(cast(int)x, cast(int)y, 4, s2, width, -rot/2+180, 0, 0xff, 0);
		if(width == 1)
		{
			int BASE = GH.SHOT_RELOAD_TIMER;
			int s = s2*(BASE-tShot)/BASE;
			if(s > 0)
			{
				Hell_drawRectEx(cast(int)x, cast(int)y, s, 4, 0, -rot/2,     0x80, 0xff, 0);
				Hell_drawRectEx(cast(int)x, cast(int)y, 4, s, 0, -rot/2+180, 0x80, 0xff, 0);
			}
		}
	}
	// 円の描画
	void drawCircle(int width, float d, float rot, int r, int g, int b)
	{
		int cnt = GH.AIM_SIZE/4;
		for(int i = 0; i < cnt; i++)
		{
//			int   r  = (isShot() || (timer%8) < 4) ? 0xff : 0x80;
			float dx = d * Hell_cosEx(rot+i*30);
			float dy = d * -Hell_sinEx(rot+i*30);
			Hell_drawRectEx(cast(int)(x+dx), cast(int)(y+dy), 4, 4, width, rot, r, g, b);
		}
	}
}

/**
 * 自機
 */
class Player : Token
{
public:
	// 状態
	enum
	{
		STATE_NORMAL, // 通常
		STATE_DAMAGE, // ダメージ
	}
	const float IMAGE_SCALE = 0.7f;
	Aim   aim;      // 照準
private:
	static bool _bKingyo;
	int   _timer;
	int   _state;
	int   _tBlowHit;
public:
	this()
	{
		aim = new Aim();
		aim.create(this, x, y);
	}
	static void setKingyo()
	{
		_bKingyo = true;
	}
	void create(float x, float y)
	{
		_timer    = 0;
		_state    = STATE_NORMAL;
		_tBlowHit = 0;
		int s     = GH.PLAYER_SIZE;
		super.create(x, y, s, s, s, s);
		_bKingyo  = false;
	}
	void vanish()
	{
		Particle.addDestroy(x, y, [0xff, 0x00, 0x00]);
		Particle.addDestroy2(x, y, [0xff, 0x00, 0x00]);
		exist = false;
	}
	bool  isDamage()   { return _state == STATE_DAMAGE; } // ダメージ中かどうか
	float getAimDegree()
	{
		float dx = aim.x - x;
		float dy = aim.y - y;
		return Hell_atan2Ex(-dy, dx);
	}
	void update()
	{
		vx = vx*GH.PLAYER_DECAY_SPEED;
		vy = vy*GH.PLAYER_DECAY_SPEED;
		switch(_state)
		{
		case STATE_NORMAL:
			update2();
			break;
		case STATE_DAMAGE:
			doDamage();
			update2();
			break;
		}
		move();
	}
	void update2()
	{
		if(_tBlowHit > 0) _tBlowHit--;
		if(_state == STATE_NORMAL)
		{
			doNormal();
			if(_timer > 0) _timer--;

		}
		if(isPushCancel())
		{
			SceneMain.addMoveCount();
			// 移動
			warp();
			Hell_stopWAV(5);
			Hell_playWAV("move", 0, 5);
			_state = STATE_NORMAL;
			vx = 0f;
			vy = 0f;
		}
		aim.update();
		rot = getAimDegree();
	}
	void update3()
	{
		if(_tBlowHit > 0) _tBlowHit--;
		if(_state == STATE_DAMAGE) _timer--;
		else                       _timer++;
	}
	void move()
	{
		if(x < w)          { x = w;          vx = -vx; }
		if(y < h)          { y = h;          vy = -vy; }
		if(x > GH.WIDTH  - w) { x = GH.WIDTH  - w; vx = -vx; }
		if(y > GH.HEIGHT - h) { y = GH.HEIGHT - h; vy = -vy; }
		super.move();
	}
	void doNormal()
	{
		_timer++;
		if(_timer > 32) _timer = 0;
		//if(aim.isShot() && isPushEnter())
		if(isPushEnter())
		{
			SceneMain.addShotCount();
			// 攻撃
			aim.shot();
		}
	}
	void doDamage()
	{
		_timer--;
		if(_timer < 1)
		{
			_state = STATE_NORMAL;
			vx = 0f;
			vy = 0f;
		}
	}
	void warp()
	{
		// ブラーエフェクト追加
		float dx = (aim.x - x)/8.0f;
		float dy = (aim.y - y)/8.0f;
		for(int i = 0; i < 8; i++) { Particle.addBlur(x+dx*i, y+dy*i, 6*(1+i), rot); }
		x = aim.x;
		y = aim.y;
	}
	bool damage(float deg, float speed)
	{
		bool ret = (_state == STATE_DAMAGE);
		if(!ret) SceneMain.addDamageCount();
		blow(deg, speed);
		_state = STATE_DAMAGE;
		_timer = GH.PLAYER_DAMAGE_TIMER;
		return ret;
	}
	void blowHit(float deg, float speed)
	{
		beginBlow();
		blow(deg, speed);
	}
	void beginBlow()
	{
		_tBlowHit = GH.PLAYER_COUNT_TIMER;
	}
	void blow(float deg, float speed)
	{
		vx = Hell_cosEx(deg)*speed;
		vy = -Hell_sinEx(deg)*speed;
	}
	void draw(bool bDanger)
	{
		if(!exist) return;

		Hell_setAlpha(HELL_ALPHA_NORMAL);
		if(_tBlowHit > 0)
		{
			drawBlowHit();
			Hell_setAlpha(HELL_ALPHA_ADD);
		}
		switch(_state)
		{
		case STATE_NORMAL:
			Hell_drawTextureEx("player", x, y,(_timer/16%2)*128, 0, 128, 128, IMAGE_SCALE, IMAGE_SCALE, rot+Hell_sinEx(_timer*11)*2);
			if(_bKingyo)
			{
				Hell_drawTextureEx("player", x, y, 3*128, 0, 128, 128, IMAGE_SCALE, IMAGE_SCALE, rot+Hell_sinEx(_timer*11)*2);
			}
			break;
		case STATE_DAMAGE:
			Hell_drawTextureEx("player", x, y, 2*128, 0, 128, 128, IMAGE_SCALE, IMAGE_SCALE, rot+Hell_sinEx(_timer*11)*2);
			break;
		}
		aim.draw(bDanger);
	}
	// 体当たりヒットの描画
	void drawBlowHit()
	{
		// 残像
		{
			Hell_setAlpha(HELL_ALPHA_NORMAL);
			float rate  = 1f * _tBlowHit/GH.PLAYER_COUNT_TIMER;
			int   alpha = cast(int)(0xff * (1-rate));
			float scale = IMAGE_SCALE + IMAGE_SCALE * rate;
			Hell_drawTextureEx("player", x, y, 128, 0, 128, 128, scale, scale, rot, 0xff, 0xff, 0xff, alpha);
		}
		// 輪
		Hell_setAlpha(HELL_ALPHA_ADD);
		for(int i = 0; i < 3; i++)
		{
			int t = _tBlowHit;
			t -= i * 10;
			if(t < 0) break;
			float radius = w + w * (GH.PLAYER_COUNT_TIMER-t)/GH.PLAYER_COUNT_TIMER;
			int   alpha  = 0xff * t/GH.PLAYER_COUNT_TIMER;
			Hell_drawCircle(x, y, radius, 3, 0xff, 0, 0, alpha);
		}
	}
}

/**
 * 自機のショット
 */
class Shot : Token
{
public:
	static int SIZE;
	static TokenManager!(Shot) parent; // 親
	int    id;
	int    timer;
	int    timer2;
	int    size;
	static void init()
	{
		SIZE = GH.SHOT_SIZE;
	}
	void create(int id, float x, float y, float deg, float speed=0, int timer=0)
	{
		this.id    = id;
		this.rot   = deg;
		this.timer = timer;
		int w      = SIZE;
		int h      = SIZE;
		super.create(x, y, w, h, w, h, Hell_cosEx(deg)*speed, -Hell_sinEx(deg)*speed);
		timer2 = 0;
	}
	static Shot add(int id, float x, float y, float deg, float speed=0, int timer=0)
	{
		Shot shot = parent.getToken();
		if(!shot) return null;
		shot.create(id, x, y, deg, speed, timer);
		return shot;
	}
	// ノーマルショット
	static void addNormal(float x, float y, float deg, float speed)
	{
		Shot shot = add(0, x, y, deg, speed, 60);
	}
	static void addNormal2(float x, float y, float ratio)
	{
		int t = cast(int)(60 * ratio / SIZE);
		int s = cast(int)(SIZE*ratio);
		Shot shot = add(1, x, y, 0, t, s);
		if(shot)
		{
			shot.w    = s;
			shot.h    = s;
			shot.size = s;
		}
	}
	float getRot(float px, float py)
	{
		// return rot;
		float dx = px - x;
		float dy = py - y;
		return Hell_atan2Ex(-dy, dx);
	}
	float getRot2(float vx, float vy)
	{
		// 反転
		float dx = -vx;
		float dy = -vy;
		return Hell_atan2Ex(-dy, dx);
	}
	void update()
	{
		switch(id)
		{
		default:
			timer = timer*GH.SHOT_DECAY/100;
			timer2++;
			if(timer < 1) exist = false;
			break;
		}
		super.move();
	}
	void draw()
	{
		Hell_setAlpha(HELL_ALPHA_ADD);
		switch(id)
		{
		case 0:
			// ノーマルショット
			float d = Hell_sinEx(180*timer/60);
			Hell_drawTextureEx("shot", x, y, 0, 0, 0, 0, 1.3f*d, 1.3f*d, timer);
			break;
		case 1:
			float d     = timer/60f;
			int   alpha = cast(int)(0xff*d);
			Hell_drawTextureEx("shot", x, y, 0, 0, 0, 0, 1.3f*d, 1.3f*d, timer*4, 0xff, 0xff, 0xff, alpha);
			if(timer2 < 6)
			{
				d += timer2/60f;
				Hell_drawTextureEx("shot", x, y, 0, 0, 0, 0, 1.3f*d, 1.3f*d, timer*4, 0x00, 0xff, 0x00, alpha);
			}
			break;
		}
	}
}

/**
 * アイテム
 */
class Item : Token
{
public:
	static TokenManager!(Item) parent; // 親
	int    id;
	int    timer;
	void create(int id, float x, float y, float deg, float speed=0, int timer=0, float ax=0, float ay=0)
	{
		this.id    = id;
		this.rot   = deg;
		this.timer = timer;
		int w      = 8;
		int h      = 8;
		super.create(x, y, w, h, w, h, Hell_cosEx(deg)*speed, -Hell_sinEx(deg)*speed, ax, ay);
	}
	static Item add(int id, float x, float y, float deg, float speed=0, int timer=0, float ax=0, float ay=0)
	{
		Item item = parent.getToken();
		if(item is null) return null;
		item.create(id, x, y, deg, speed, timer, ax, ay);
		return item;
	}
	// 自由落下バナナの追加
	static void addBananaEx(float x, float y, int cnt)
	{
		for(int i = 0; i < cnt; i++)
		{
			float deg   = Hell_randint(60, 120);
			float speed = 5 + Hell_randint(0, 100)/100f;
			Item banana = add(1, x, y, deg, speed, 9999+Hell_randint(0, 359), 0, 0.01f);
		}
	}
	// 自由落下バナナ２の追加
	static void addBanana2Ex(float x, float y, int cnt)
	{
		for(int i = 0; i < cnt; i++)
		{
			float deg   = Hell_randint(60, 120);
			float speed = 5 + Hell_randint(0, 100)/100f;
			Item banana = add(2, x, y, deg, speed, 9999+Hell_randint(0, 359), 0, 0.01f);
		}
	}
	void vanish()
	{
		Particle.addDestroy(x, y, [0xff, 0xff, 0x00]);
		exist = false;
	}
	void update()
	{
		switch(id)
		{
		case 1:
		case 2:
			timer--;
			if(timer < 1) exist = false;
			if(vy < -1) vy = -1;
			if(vy > 1)  vy = 1;
			vx = vx * 0.97f;
			break;
		default:
			timer--;
			if(timer < 1) exist = false;
			break;
		}
		super.move();
		if(isOutOfWindow2()) exist = false;
	}
	void draw()
	{
		int alpha = 0xff;
		switch(id)
		{
		case 1:
			// バナナ
			Hell_setAlpha(HELL_ALPHA_NORMAL);
			Hell_drawTextureEx("enemy", x, y, 256, 64, 64, 64, 1f, 1f, timer*2);
			alpha = cast(int)(0x80 * Hell_sinEx(timer*4%180));
			Hell_drawTextureEx("enemy", x, y, 320, 0, 64, 64, 1f, 1f, timer*2, 0xff, 0xff, 0xff, alpha);
			break;
		case 2:
			// バナナ２
			Hell_setAlpha(HELL_ALPHA_NORMAL);
			Hell_drawTextureEx("enemy", x, y, 256, 64, 64, 64, .5f, .5f, timer*2);
			alpha = cast(int)(0x80 * Hell_sinEx(timer*4%180));
			Hell_drawTextureEx("enemy", x, y, 320, 0, 64, 64, .5f, .5f, timer*2, 0xff, 0xff, 0xff, alpha);
			break;
		}
	}
}


/**
 * 敵弾
 */
class Bullet : Token
{
public:
	static TokenManager!(Bullet) parent; // 親
	static Player                target; // プレイヤー
	void create(float x, float y, float deg=0, float speed=0, float ax=0, float ay=0)
	{
		int s = GH.BULLET_SIZE;
		super.create(x, y, s, s, s, s, Hell_cosEx(deg)*speed, -Hell_sinEx(deg)*speed, ax, ay);
	}
	void vanish()
	{
		exist = false;
	}
	static void vanishAll()
	{
		foreach(i, b; parent.getArray())
		{
			if(!b.exist) continue;
			float dx    = GH.WIDTH/2  - b.x;
			float dy    = GH.HEIGHT/2 - b.y;
			float deg   = Hell_atan2Ex(-dy, dx);
			int   rate  = SceneMain.Ratio;
			int   score = 10*(i+1) * rate;
			Particle.addNumber(b.x, b.y, deg, 5, score, rate);
			SceneMain.Score += score;
			Particle.addDestroy(b.x, b.y, [0xff, 0x00, 0x00]);
			b.vanish();
		}
	}
	static Bullet add(float x, float y, float deg=0, float speed=0, float ax=0, float ay=0)
	{
		Bullet bullet = parent.getToken();
		if(!bullet) return null;
		bullet.create(x, y, deg, speed, ax, ay);
		return bullet;
	}
	static float getAim(float x, float y)
	{
		float dx = target.x - x;
		float dy = target.y - y;
		return Hell_atan2Ex(-dy, dx);
	}
	// 狙い撃ち弾
	static void addAim(float x, float y, float speed=3f)
	{
		return add(x, y, getAim(x, y), speed);
	}
	// NWay弾
	static void addNWay(float x, float y, float speed=3f, float cnt=3, float deg1=-5f, float deg2=5f)
	{
		float rot = getAim(x, y) + deg1;
		float d   = (deg2 - deg1)/cnt;
		for(int i = 0; i < cnt; i++)
		{
			add(x, y, rot, speed);
			rot += d;
		}
	}
	// 狙い撃ち全方向弾
	static void addAllRange(float x, float y, float speed=3f, float cnt=8, float deg=0)
	{
		float d   = 360f/cnt;
		float rot = getAim(x, y) + deg;
		for(int i = 0; i < cnt; i++)
		{
			add(x, y, rot, speed);
			rot += d;
		}
	}
	void update()
	{
		vx += ax; vy += ay;
		x  += vx; y  += vy;
		if(isOutOfWindow2()) exist = false;
	}
	void draw()
	{
		Hell_setAlpha(HELL_ALPHA_NORMAL);
		Hell_drawTextureEx("enemy", x, y, 256, 0, 64, 64);
	}
}

/**
 * 打ち返し弾
 */
class Reflection : Token
{
public:
	static TokenManager!(Reflection) parent; // 親
	int   timer;
	int   rate;
	float rotSpeed;
	void create(float x, float y, float deg=0, float speed=0)
	{
		int s = GH.REFLECT_SIZE;
		super.create(x, y, s, s, s, s, Hell_cosEx(deg)*speed, -Hell_sinEx(deg)*speed);
		rot      = cast(int)deg;
		rotSpeed = GH.REFLECT_ROT_INIT;
		timer    = 0;
	}
	void vanish()
	{
		Particle.addReflection(x, y);
		exist = false;
	}
	static Reflection add(float x, float y, int rate, float deg=0, float speed=0)
	{
		Reflection ref = parent.getToken();
		if(!ref) return null;
		ref.create(x, y, deg, speed);
		ref.rot  = cast(int)deg;
		ref.rate = rate;
		return ref;
	}
	void lookEnemy(Enemy e)
	{
		rotSpeed += GH.REFLECT_ROT_INC;
		if(rotSpeed > GH.REFLECT_ROT_MAX) rotSpeed = GH.REFLECT_ROT_MAX;
		float dx  = e.x - x;
		float dy  = e.y - y;
		float deg = Hell_atan2Ex(-dy, dx);
		float r   = deg - rot;
		if(Hell_abs(r) < 3)
		{
			rot = deg;
		}
		else
		{
			if(Hell_abs(r) > 180)
			{
				if(rot > deg) { rot -= 360; }
				else          { rot += 360; }
			}
			rot = rot*(1 - rotSpeed) + deg*rotSpeed;
		}
	}
	void update()
	{
		timer++;
		bool bExist = false; // ターゲットが存在するかどうか
		Enemy boss = Enemy.getBoss();
		if(boss is null || !boss.isHitReflect() || !GH.REFLECT_BOSS_GUIDE)
		{
			float len = GH.WIDTH*GH.WIDTH + GH.HEIGHT*GH.HEIGHT;
			Enemy target = null;
			foreach(e; Enemy.parent.getArray())
			{
				if(!e.exist)          continue;
				if(!e.isHitReflect()) continue; // 打ち返しは当たらない
				float dx = e.x - x;
				float dy = e.y - y;
				float d  = Hell_distance(dx, dy);
				if(d < len)
				{
					// ターゲット更新
					len    = d;
					target = e;
				}
			}
			if(target) lookEnemy(target);
			bExist = true;
		}
		else
		{
			lookEnemy(boss);
			bExist = true;
		}
		if(!bExist) rot += 2;
		vx = GH.REFLECT_SPEED * Hell_cosEx(rot);
		vy = GH.REFLECT_SPEED * -Hell_sinEx(rot);
		super.move();
		int margin = GH.REFLECT_MARGIN;
		int x1 = -margin;
		int y1 = -margin;
		int x2 = GH.WIDTH  + margin;
		int y2 = GH.HEIGHT + margin;
		if(x < x1 || y < y1 || x > x2 || y > y2) exist = false;
		//if(isOutOfWindow2()) exist = false;
	}
	void draw()
	{
		Hell_setAlpha(HELL_ALPHA_NORMAL);
		float scale = 1.0f;
		switch(rate)
		{
		case 1:  scale = 1.0f; break;
		case 2:  scale = 1.2f; break;
		case 4:  scale = 1.4f; break;
		case 8:  scale = 1.6f; break;
		case 16: scale = 1.8f; break;
		case 32: scale = 2.0f; break;
		default: break;
		}
		Hell_drawTextureEx("enemy", x, y, 384+32*((timer%80)/20), 256-32, 32, 32, scale, scale, rot);
	}
}

class Back
{
public:
	enum
	{
		STATE_NONE,
		STATE_NORMAL,
		STATE_RASTER,
		STATE_ROT,
		STATE_BLACK_OUT,
		STATE_NOISE,
	}
	int timer;
	int state;
public:
	this() { init(); }
	void init()
	{
		timer = 0;
		state = STATE_NORMAL;
	}
	void begin(int state, int timer)
	{
		this.state = state;
		this.timer = timer;
	}
	void update()
	{
		switch(state)
		{
		case STATE_NORMAL:
			break;
		case STATE_NONE:
		case STATE_RASTER:
		case STATE_NOISE:
			timer--;
			if(timer == 0) state = STATE_NORMAL;
			break;
			break;
		case STATE_ROT:
			timer = timer * 97 / 100;
			if(timer < 1) state = STATE_NORMAL;
			break;
		case STATE_BLACK_OUT:
			if(timer > 0) timer--;
			break;
		}
	}
	void draw()
	{
		Hell_setAlpha(HELL_ALPHA_NORMAL);
		switch(state)
		{
		case STATE_NONE:
			// 何も描画しない
			break;
		case STATE_NORMAL:
			float scale = 1.f;
			Hell_drawTextureEx("back", 320, 240, 0, 0, GH.WIDTH, GH.HEIGHT, scale, scale);
			break;
		case STATE_RASTER:
			for(int i = 0; i < GH.HEIGHT; i++)
			{
				const int D = 12;
				float r = (1f*timer)/GH.MAIN_SUICIDE_TIMER;
				int t = timer*D + cast(int)(D*i*r);
				if(i%2 == 0) t = -t;
				int x = cast(int)(D*10*Hell_sinEx(t));
				int y = i;
				int c = 0xff-cast(int)(0x80*r);
				Hell_drawTexture("back", x, y, 0, y, GH.WIDTH, 1, 1f, 1f, 0f, c, c, 0xff);
			}
			break;
		case STATE_ROT:
			Hell_drawTextureEx("back", 320, 240, 0, 0, GH.WIDTH, GH.HEIGHT, 1f, 1f, -timer*4, 0xff, 0xff, 0xff, 0xa0);
			Hell_drawTextureEx("back", 320, 240, 0, 0, GH.WIDTH, GH.HEIGHT, 1f, 1f, timer*4,  0xff, 0xff, 0xff, 0xa0);
			break;
		case STATE_BLACK_OUT:
			int alpha = timer+0x80;
			if(timer > 0)
			{
				if(alpha > 0xff) alpha = 0xff;
				Hell_drawTextureEx("back", 320, 240, 0, 0, GH.WIDTH, GH.HEIGHT, 1f, 1f, 0, 0xff, 0xff, 0xff, alpha);
			}
			else
			{
				alpha = 0x80;
				if(Hell_randint(0, 100) == 0) alpha = 0xc0;
				Hell_drawTextureEx("back", 320, 240, 0, 0, GH.WIDTH, GH.HEIGHT, 1f, 1f, 0, 0xff, 0xff, 0xff, alpha);
			}
			break;
		case STATE_NOISE:
			for(int i = 0; i < GH.HEIGHT; i++)
			{
				int x = Hell_randint(-timer, timer);
				int y = i;
				Hell_drawTexture("back", x, y, 0, y, GH.WIDTH, 1, 1f, 1f, 0f);
			}
			break;
		}
	}
}

/**
 * メインシーン
 */
class SceneMain
{
public:
	// 定数
	const int TIMER_INIT           = 120;
	const int TIMER_GAMEOVER_INIT  = 60;
	const int TIMER_GAMECLEAR_INIT = 60;
	const int TIMER_GAMECLEAR      = 90;
	const int TIMER_RECORD_SMASH   = 40;
	const int TIMER_SCORE_INIT     = 180;
	const int TIMER_SCORE_MAIN     = 10;
	const int MODE_H               = 40;
	const int MODE_X               = 200;
	const int MODE1_Y              = 220;
	const int MODE2_Y              = 220 + MODE_H;
	const int MODE3_Y              = 220 + MODE_H*2;
	// 状態
	enum {
		STATE_INIT,           // 初期化
		STATE_MAIN,           // メイン
		STATE_HIT_WAIT,       // 敵ヒット
		STATE_DAMAGE_WAIT,    // 自機ダメージ
		STATE_GAMEOVER_INIT,  // ゲームオーバー(初期化)
		STATE_GAMEOVER,       // ゲームオーバー
		STATE_BOSS_SUICIDE,   // ボス自爆中
		STATE_GAMECLEAR_INIT, // ゲームクリアー(初期化)
		STATE_GAMECLEAR,      // ゲームクリアー
		STATE_SCORE_INIT,
		STATE_SCORE_MAIN,
		STATE_MENU,           // メニュー
	};
	// スタティック変数
	static int   tLimit;           // 制限時間
	static int   tLimitMax;        // 制限時間の最大値
	static int[] tBanana;          // お助けバナナ
	static bool  bEndAtAllDestroy; // 敵全滅でゲーム終了するかどうか
	static int   dmg1Val;          // ダメージフレームの減少時間
	static int   dmg2Val;          // ダメージ中の減少時間
	static int   dmg2Interval;     // ダメージ中の減少時間の間隔
	static int   Ratio;            // スコア倍率
	static int   Score;            // スコア
	static Back  back;             // 背景管理
	// スコア
	static int   cntMove;          // 移動回数
	static int   cntShot;          // ショット回数
	static int   cntChain;         // 連鎖回数
	static int   cntMissile;       // ミサイル発射数
	static int   cntDamage;        // ダメージ回数
	
	// 変数
	LevelManager level;           // レベル管理
	Player       player;          // プレイヤー
	Message      message;         // メッセージ
	ConfigParser savedata;        // セーブデータ
	int          state;           // 状態
	int          timer;           // タイマ
	int          selected;        // メニュー選択項目
	int          tMin;            // 分
	int          tSec;            // 秒
	int          tMSec;           // ミリ秒
	int          tSecPrev;        // 秒(１フレーム前)
	int          tPast;           // 経過時間
	int          tBest;           // 最短クリア時間
	int          tRecordSmash;    // 記録更新タイマ
	int          tInit;           // 初期化タイマ
	int          tCount;          // 汎用カウンタ
	int scoreMove;        // 移動のスコア
	int scoreShot;        // ショットのスコア
	int scoreChain;       // 連鎖回数のスコア
	int scoreMissile;     // ミサイル発射のスコア
	int scoreLastTime;    // 残り時間のスコア
	int scoreClearTime;   // クリア時間のスコア
	int scoreDamage;      // ダメージ回数のスコア
	int scoreTotal;       // トータルスコア
	int scorePrev;        // 現在のベストスコア
	int gradeNow;         // 現在の優良スコアの数
	int gradePrev;        // ベスト優良スコアの数
public:
	this()
	{
		savedata          = new ConfigParser();
		level             = new LevelManager();
		
		player            = new Player();
		message           = new Message();
		back              = new Back();
		Shot.parent       = new TokenManager!(Shot)(32);
		Reflection.parent = new TokenManager!(Reflection)(512);
		Item.parent       = new TokenManager!(Item)(32);
		Enemy.parent      = new TokenManager!(Enemy)(64);
		Bullet.parent     = new TokenManager!(Bullet)(256);
		Particle.parent   = new TokenManager!(Particle)(512);
		Enemy.target      = player;
		Bullet.target     = player;
		Shot.init();
	}
	void calcScore()
	{
		// ショット回数
		if(cntShot > 100) scoreShot = 0;
		else
		{
			scoreShot = ((100 - cntShot) / 10 + 1) * 1000;
		}
		// 連鎖回数
		scoreChain = cntChain * 1000;
		// ミサイル
		scoreMissile = cntMissile * 10;
		// 移動回数
		if(cntMove > 100) scoreMove = 0;
		else
		{
			scoreMove = ((100 - cntMove) / 10 + 1) * 1000;
		}
		// 残り時間
		scoreLastTime = (10000 * tLimit / tLimitMax / 10) * 10;
		// クリア時間
		int sec = tPast/60;
		int score = 22000;
		scoreClearTime = ((score - sec*100) / 10) * 10;
		if(scoreClearTime < 0) scoreClearTime = 0;
		// ダメージ回数
		scoreDamage = 13000 - 1000 * cntDamage;
		if(scoreDamage < 0) scoreDamage = 0;
		// トータルスコア
		scoreTotal = scoreShot + scoreMove + scoreChain + scoreMissile + scoreLastTime + scoreClearTime + scoreDamage;
		
		gradeNow = 0;
		if(scoreShot      >= 8000) gradeNow++;
		if(scoreMove      >= 8000) gradeNow++;
		if(scoreChain     >= 8000) gradeNow++;
		if(scoreMissile   >= 8000) gradeNow++;
		if(scoreLastTime  >= 8000) gradeNow++;
		if(scoreClearTime >= 8000) gradeNow++;
		if(scoreDamage    >= 8000) gradeNow++;
	}
	// 初期化
	void init()
	{
		savedata.load(GH.SAVE_DATA);
		Hell_setBgmVolume(128);
		switch(gMode)
		{
		case 0:
			level.load("resource/level/easy");
			tBest = savedata.getint("EASY");
			scorePrev = savedata.getint("EASY_SCORE");
			gradePrev = savedata.getint("EASY_GRADE");
			break;
		case 1:
			level.load("resource/level/normal");
			tBest = savedata.getint("NORMAL");
			scorePrev = savedata.getint("NORMAL_SCORE");
			gradePrev = savedata.getint("NORMAL_GRADE");
			break;
		case 2:
			level.load("resource/level/hard");
			tBest = savedata.getint("HARD");
			scorePrev = savedata.getint("HARD_SCORE");
			gradePrev = savedata.getint("HARD_GRADE");
			break;
		case 3:
			level.load("resource/level/hell");
			tBest = savedata.getint("HELL");
			scorePrev = savedata.getint("HELL_SCORE");
			gradePrev = savedata.getint("HELL_GRADE");
			break;
		default:
			throw new Error("TODO!!!");
			break;
		}
		// トークン初期化
		Shot.parent.vanish();
		Reflection.parent.vanish();
		Item.parent.vanish();
		Enemy.parent.vanish();
		Bullet.parent.vanish();
		Particle.parent.vanish();
		
		player.create(0, 600);
		message.init();
		message.beginSomething();
		back.init();
		state    = STATE_INIT;
		timer    = 0;
		selected = 0;
		setLimitTime(3*60); // デフォルトは3min
		bEndAtAllDestroy = false;
		g_fps = 1000/60;
		tBanana.length = 0;
		Ratio  = 1;
		Score = 0;
		tPast = 0;
		tRecordSmash = 0;
		tCount = 0;
		// スコア
		cntMove    = 0; // 移動回数
		cntShot    = 0; // ショット回数
		cntChain   = 0; // 最大連鎖回数
		cntMissile = 0; // ミサイル発射数
		cntDamage  = 0; // ダメージ回数
		
		tInit = TIMER_INIT;
		Hell_playWAV("pi");
	}
	static void addMoveCount()           { cntMove++;  }
	static void addShotCount()           { cntShot++;  }
	static void setChainCount(int val)   { if(val > cntChain) cntChain = val; }
	static void addMissileCount(int val) { cntMissile += val; }
	static void addDamageCount()         { cntDamage++; }
	// 敵全滅でゲーム終了
	static void beginEndAtAllDestroy()        { bEndAtAllDestroy = true; }
	// ダメージフレームの減少時間の設定
	static void setDmg1Val(int sec)           { dmg1Val = sec; }
	// ダメージ中の減少時間
	static void setDmg2Val(int sec)           { dmg2Val = sec; }
	// ダメージ中の減少時間の間隔
	static void setDmg2Interval(int interval) { dmg2Interval = interval; }
	// お助けバナナの出現時間
	static void setBanana(int sec)            { tBanana ~= sec*60; }
	// 背景エフェクト開始
	static void setBack(int arg1, int arg2)   { back.begin(arg1, arg2); }
	
	bool isGameClear()
	{
		if(!bEndAtAllDestroy)        return false; // 全滅クリア条件でない
		if(Enemy.parent.count() > 0) return false; // 敵がまだ残っている
		return true;
	}
	void update()
	{
		tCount++;
		// メニュー表示時は更新しない
		if(state != STATE_MENU) { Particle.parent.update(); }
		switch(state)
		{
		case STATE_MAIN:
		case STATE_HIT_WAIT:
		case STATE_DAMAGE_WAIT:
			tPast++;
			if(tPast > 59*60*60 + 59*60 + 59) tPast = 59*60*60 + 59*60 + 59;
			break;
		default:
			break;
		}
		
		message.update();
		back.update();
		switch(state)
		{
		case STATE_INIT:
			player.rot += tInit;
			player.x = GH.WIDTH/2  - 1*tInit;
			player.y = GH.HEIGHT/2 + (GH.HEIGHT/2)*tInit/TIMER_INIT;
			if(tInit > 0) tInit = tInit*90/100;
			if(tInit == 0) state = STATE_MAIN;
			break;
		case STATE_MAIN:
			if(isGameClear())
			{
				Item.parent.vanish();
				// ゲームクリアー
				if(tPast < tBest)
				{
					// 最短時間更新
					switch(gMode)
					{
					case 0: savedata.set("EASY",   cast(int)tPast); break;
					case 1: savedata.set("NORMAL", cast(int)tPast); break;
					case 2: savedata.set("HARD",   cast(int)tPast); break;
					case 3: savedata.set("HELL",   cast(int)tPast); break;
					default:
						throw new Error("TODO!!!");
						break;
					}
					savedata.write();
					tRecordSmash = TIMER_RECORD_SMASH;
				}
				// クリア
				if(gMode == 3)
				{
					savedata.set("HELL_CLEAR", true);
					savedata.write();
				}
				
				if(Hell_getBgmVolume() < 128)
				{
					Hell_setBgmVolume(128);
					Hell_wait(100);
				}
				g_fps = 1000/15;
				state = STATE_GAMECLEAR_INIT;
				timer = TIMER_GAMECLEAR_INIT;
				return;
			}
			updateLimit(tLimit, tLimit-1);
			if(isTimeOver())
			{
				// 時間切れ
				Hell_stopBgm(1000);
				player.damage(0, 0);
				state = STATE_GAMEOVER_INIT;
				timer = TIMER_GAMEOVER_INIT;
				return;
			}
			if(isPushMenu())
			{
				// メニューを開く
				selected = 2;
				state    = STATE_MENU;
				return;
			}
			doMain();
			break;
		case STATE_HIT_WAIT:
		case STATE_DAMAGE_WAIT:
			player.update2();
			timer--;
			if(timer < 1)
			{
				state = STATE_MAIN;
			}
			break;
		case STATE_BOSS_SUICIDE:
			Enemy.updateBoss();
			timer--;
			if(timer < 1)
			{
				state = STATE_MAIN;
			}
			break;
		case STATE_GAMEOVER_INIT:
			timer--;
			if(timer < 1)
			{
				Hell_playWAV("destroy1");
				player.vanish();
				state = STATE_GAMEOVER;
			}
			break;
		case STATE_GAMEOVER:
			if(isPushEnter())
			{
				gotoTitle();
			}
			break;
		case STATE_GAMECLEAR_INIT:
			int t    = timer/2;
			int ofsX = timer%2 == 0 ? -t : t;
			Hell_setOffset(ofsX, Hell_randint(-t, t));
			player.update3();
			Shot.parent.update();
			Reflection.parent.update();
			Item.parent.update();
			timer--;
			if(isPushMenu())
			{
				// スキップ
				timer = 0;
			}
			if(timer < 1)
			{
				for(int i = 0; i < 8; i++)
				{
					int   x     = Hell_randint(GH.WIDTH/8,  GH.WIDTH*7/8);
					int   y     = Hell_randint(GH.HEIGHT/8, GH.HEIGHT*7/8);
					int   deg   = Hell_randint(0, 359);
					float speed = Hell_randint(10, 40)/10f;
					Particle.addFlare(x, y, deg, speed);
				}
				Hell_setOffset(0, 0);
				message.beginSomething();
				Hell_stopBgm(5000);
				g_fps = 1000/60;
				Hell_playWAV("clear");
				state = STATE_GAMECLEAR;
				timer = TIMER_GAMECLEAR;
			}
			break;
		case STATE_GAMECLEAR:
			if(tRecordSmash > 0) tRecordSmash--;
			if(timer > 0)
			{
				timer--;
				if(isPushEnter())
				{
					timer = 0;
				}
				if(timer == 0) Hell_playWAV("kin");
			}
			else
			{
				tRecordSmash = 0;
				if(isPushEnter())
				{
					calcScore();
					if(scoreTotal > scorePrev)
					{
						// スコア更新
						switch(gMode)
						{
						case 0: savedata.set("EASY_SCORE",   scoreTotal); break;
						case 1: savedata.set("NORMAL_SCORE", scoreTotal); break;
						case 2: savedata.set("HARD_SCORE",   scoreTotal); break;
						case 3: savedata.set("HELL_SCORE",   scoreTotal); break;
						default:
							throw new Error("TODO!!!");
							break;
						}
						savedata.write();
					}
					if(gradeNow > gradePrev)
					{
						// スコア更新
						switch(gMode)
						{
						case 0: savedata.set("EASY_GRADE",   gradeNow); break;
						case 1: savedata.set("NORMAL_GRADE", gradeNow); break;
						case 2: savedata.set("HARD_GRADE",   gradeNow); break;
						case 3: savedata.set("HELL_GRADE",   gradeNow); break;
						default:
							throw new Error("TODO!!!");
							break;
						}
						savedata.write();
					}
					timer = TIMER_SCORE_INIT;
					state = STATE_SCORE_INIT;
					Hell_playBgm("resource/sound/10.ogg", 0);
				}
			}
			break;
		case STATE_SCORE_INIT:
			timer--;
			if(timer < 1)
			{
				timer = TIMER_SCORE_MAIN;
				state = STATE_SCORE_MAIN;
			}
			break;
		case STATE_SCORE_MAIN:
			if(timer > 0)
			{
				timer--;
			}
			else
			{
				if(isPushEnter()) gotoTitle();
			}
			break;
		case STATE_MENU:
			timer++;
			player.aim.updateAnime();
			int my = Hell_getMouseY();
			if(MODE1_Y <= my && my < MODE1_Y + MODE_H) selected = 0;
			if(MODE2_Y <= my && my < MODE2_Y + MODE_H) selected = 1;
			if(MODE3_Y <= my && my < MODE3_Y + MODE_H) selected = 2;
			if(isPushEnter())
			{
				switch(selected)
				{
				case 0:
					init();
					Hell_playBgm("resource/sound/5.ogg");
					break;
				case 1:
					gotoTitle();
					break;
				case 2:
					state = STATE_MAIN;
					break;
				}
				return;
			}
			if(isPushMenu())
			{
				state = STATE_MAIN;
			}
			break;
		}
	}
	// 制限時間の設定
	static void setLimitTime(int sec)
	{
		tLimit    = sec*60; // 60fps
		tLimitMax = tLimit;
	}
	// 制限時間のエクステンド
	void addLimitTime(int sec)
	{
		updateLimit(tLimit, tLimit + sec*60); // 60fps
	}
	void damagePlayer()
	{
		if(dmg1Val <= 0) return;

		Particle.addMessage(player.x, player.y, format("-%d", dmg1Val), [0xff, 0x00, 0x00]);
		addLimitTime(-dmg1Val);
	}
	// プレイヤーダメージ時の判定
	void updatePlayerDamage()
	{
		if(!player.isDamage())      return; // ダメージ中でない
		if(dmg2Interval <= 0)       return; // 減少タイマなし
		if(tLimit%dmg2Interval > 0) return; // 減少するタイミングでない
		if(dmg2Val <= 0)            return; // ダメージなし
		// 時間減少
		Particle.addMessage(player.x, player.y, format("-%d", dmg2Val), [0xff, 0x00, 0x00]);
		addLimitTime(-dmg2Val);
	}
	// 制限時間の更新
	void updateLimit(int prev, int now)
	{
		tLimit = now;
		if(tLimit > tLimitMax) tLimit = tLimitMax;
		if(tLimit < 0)         tLimit = 0;
		foreach(t; tBanana)
		{
			if(prev > t && now <= t)
			{
				Enemy e = Enemy.addEx(1, Hell_randint(GH.WIDTH/3, 2*GH.WIDTH/3), Hell_randint(GH.HEIGHT/3, 2*GH.HEIGHT/3));
				e.setAttribute(1);
				return;
			}
		}
		tSecPrev = tSec;
		tMin     = tLimit/60/60;
		tSec     = (tLimit/60)%60;
		tMSec    = 1000*(tLimit%60)/60;
		if(tMin < 1 && tSec < 10)
		{
			if(Hell_getBgmVolume() > 32)
			{
				Hell_setBgmVolume(32);
				Hell_wait(100);
			}
		}
		else
		{
			if(Hell_getBgmVolume() < 128)
			{
				Hell_setBgmVolume(128);
				Hell_wait(100);
			}
		}
		if(tMin < 1 && tSec < 10 && tSec < tSecPrev)
		{
			Hell_playWAV("clock");
			int sec = tSec+1;
			if(sec > 0) Hell_playWAV(format("v%d", sec));
		}
	}
	// 制限時間切れ間近
	bool isDanger()
	{
		return tLimit <= 10 * 60; // 10sec
	}
	void setRatio(int val)
	{
		Ratio = val;
	}
	void updateRatio(int val)
	{
		if(val < Ratio) return;
		if(val > GH.MAIN_RATIO_MAX) Ratio = GH.MAIN_RATIO_MAX;
		else                        Ratio = val;
	}
	bool isTimeOver()
	{
		return tLimit <= 0;
	}
	void doMain()
	{
		level.update();
		player.update();
		if(isDanger()) player.aim.updateAnime();
		Shot.parent.update();
		Reflection.parent.update();
		Item.parent.update();
		Enemy.parent.update();
		Bullet.parent.update();
		updatePlayerDamage(); // ダメージの更新
		
		if(Reflection.parent.count() == 0) setRatio(1);
		
		bool bSndEnemy  = false; // 敵撃破
		bool bSndEnemy2 = false; // 敵体当たり撃破
		bool bSndRefHit = false; // 打ち返しヒット
		bool bSndItem   = false; // アイテム取得
		// 当たり判定
		// 弾
		foreach(bullet; Bullet.parent.getArray())
		{
			if(!bullet.exist) continue;
			// vsショット
			foreach(shot; Shot.parent.getArray())
			{
				if(!shot.exist) continue;
				if(shot.isHit2(bullet))
				{
					Particle.addDestroy3(bullet.x, bullet.y, [0xff, 0, 0]);
					bullet.vanish();
				}
			}
			// vs照準
			if(bullet.isHit2(player.aim))
			{
				if(isDanger() || GH.AIM_REFLECT)
				{
					// 弾消しで打ち返し
					//if(player.aim.isShot())
					{
						float dx  = bullet.x - player.aim.x;
						float dy  = bullet.y - player.aim.y;
						float rot = Hell_atan2Ex(-dy, dx);
						Reflection.add(bullet.x, bullet.y, 1, rot, GH.REFLECT_SPEED);
					}
					SceneMain.addMissileCount(1);
				}
				Particle.addDestroy3(bullet.x, bullet.y, [0xff, 0, 0]);
				bullet.vanish();
				continue;
			}
			// vs自機
			if(bullet.isHit2(player))
			{
				float dx  = player.x - bullet.x;
				float dy  = player.y - bullet.y;
				float rot = Hell_atan2Ex(-dy, dx);
				if(!player.damage(rot, GH.PLAYER_DAMAGE_SPEED))
				{
					damagePlayer();
					state = STATE_DAMAGE_WAIT;
					timer = GH.MAIN_DAMAGE_TIMER;
					return;
				}
			}

		}
		// アイテムvs自機(照準)
		foreach(item; Item.parent.getArray())
		{
			if(!item.exist) continue;
			if(item.isHit2(player) || item.isHit2(player.aim))
			{
				switch(item.id)
				{
				case 1:
					// バナナ取得
					addLimitTime(GH.ITEM_BANANA_POWER);
					Particle.addMessage(item.x, item.y, format("+%d", GH.ITEM_BANANA_POWER), [0xff, 0xff, 0x00]);
					break;
				case 2:
					// バナナ２取得
					addLimitTime(GH.ITEM_BANANA2_POWER);
					Particle.addMessage(item.x, item.y, format("+%d", GH.ITEM_BANANA2_POWER), [0xff, 0xff, 0x00]);
					break;
				}
				item.vanish();
				bSndItem = true;
				continue;
			}
			float dx = player.x - item.x;
			float dy = player.y - item.y;
			if(dx*dx + dy*dy < GH.ITEM_SORB_SIZE*GH.ITEM_SORB_SIZE)
			{
				// 
				float rad = Hell_atan2(-dy, dx);
				item.vx = GH.ITEM_SORB_SPEED*Hell_cos(rad);
				item.vy = GH.ITEM_SORB_SPEED*-Hell_sin(rad);
			}
		}
		// 打ち返しvs敵
		foreach(ref; Reflection.parent.getArray())
		{
			if(!ref.exist) continue;
			foreach(enemy; Enemy.parent.getArray())
			{
				if(!enemy.exist)          continue;
				if(!enemy.isHitReflect()) continue; // 打ち返しは当たらない
				if(!enemy.isHit2(ref))    continue;
				// 打ち返しヒット
				enemy.damageEx(ref.rot, GH.REFLECT_BLOW_SPEED, GH.REFLECT_DAMAGE);
				ref.vanish();
				bSndRefHit = true;
				if(enemy.isDead())
				{
					enemy.ratio = Ratio;
					int score = enemy.vanish();
					bSndEnemy = true;
					state = STATE_HIT_WAIT;
					timer = GH.MAIN_ATTACK_TIMER;
				}
				else
				{
					state = STATE_HIT_WAIT;
					timer = GH.MAIN_REFLECT_TIMER;
				}
				break;
			}
		}
		// ショットvs敵
		foreach(shot; Shot.parent.getArray())
		{
			if(!shot.exist) continue;
			foreach(enemy; Enemy.parent.getArray())
			{
				if(!enemy.exist)        continue;
				if(enemy.isBarrier())   continue; // バリアがあるのでショットは効かない
				if(!enemy.isHit2(shot)) continue; // 当たっていない
				enemy.damage(shot.getRot(enemy.x, enemy.y), GH.SHOT_BLOW_SPEED, 0);
			}
		}
		// 敵vs敵
		foreach(e1; Enemy.parent.getArray())
		{
			if(!e1.exist)        continue;
			if(!e1.isDamage())   continue; // ダメージ中でない
			// ダメージ中の敵と...
			foreach(e2; Enemy.parent.getArray())
			{
				if(!e2.exist)       continue;
				if(!e2.isHitShot()) continue; // ショットは当たらない
				if(e2.isDamage())   continue; // ダメージ中
				// ダメージ中でない敵が...
				float w1  = e1.w * GH.ENEMY_DAMAGE_SIZE; // ダメージ中の当たり判定は「ENEMY_DAMAGE_SIZE」倍
				if(isHitCircle(e1.x, e1.y, e2.x, e2.y, w1, e2.w))
				{
					// 衝突
					// 当たったほうが消滅
					int tmp = e1.ratio * 2;
					if(tmp > GH.MAIN_RATIO_MAX) tmp = GH.MAIN_RATIO_MAX;
					if(tmp >= 4)
					{
						// 連鎖バナナ生成
						Item.addBanana2Ex(e1.x, e1.y, 1);
					}
					e2.ratio = tmp;
					SceneMain.setChainCount(e1.chain);
					e2.chain = e1.chain + 1;
					message.beginCombo(e2.ratio);
					float dx = e2.x - e1.x;
					float dy = e2.y - e1.y;
					float rot = Hell_atan2Ex(-dy, dx);
					float mag = Hell_log10(e1.ratio*10);
					int   val = cast(int)(GH.ENEMY_BASE_DAMAGE*(mag*mag));
					e2.damage(rot, GH.ENEMY_BLOW_SPEED, val);
					int score = e1.vanish();
					bSndEnemy = true;
					timer = GH.MAIN_ATTACK_TIMER;
					state = STATE_HIT_WAIT;
				}
				
			}
		}
		// 敵vs自機
		foreach(enemy; Enemy.parent.getArray())
		{
			if(!enemy.exist)        continue;
			if(enemy.isDamage())
			{
				float w1  = enemy.w * GH.ENEMY_DAMAGE_SIZE; // ダメージ中は大きめで
				if(!isHitCircle(player.x, player.y, enemy.x, enemy.y, player.w, w1)) continue; // 当たっていない
				
				// ダメージ中の敵と衝突
				// ダメージ中の敵は倒せる
				updateRatio(enemy.ratio); // 最大倍率更新
				int score = enemy.vanish();
				bSndEnemy  = true;
				bSndEnemy2 = true;
				float dx = player.x - enemy.x;
				float dy = player.y - enemy.y;
				float rot = Hell_atan2Ex(-dy, dx);
				// 打ち返し弾発動
				int cnt = enemy.getReflectCount();
				message.beginReflect(cnt);
				SceneMain.addMissileCount(cnt);
				float d = 0f;
				if(cnt > 0) { d = 360.0f / cnt; }
				float deg = Hell_randint(0, 359);
				for(int i = 0; i < cnt; i++)
				{
					Reflection.add(player.x, player.y, enemy.ratio, deg, GH.REFLECT_SPEED);
					deg += d;
				}
				Particle.addCircle(player.x, player.y, [0xff, 0, 0]);
				player.blowHit(rot, GH.PLAYER_DAMAGE_SPEED);
				state = STATE_HIT_WAIT;
				timer = GH.MAIN_ATTACK_TIMER;
				continue;
			}
			if(!enemy.isHitShot())    continue; // ショットが当たらない
			if(!enemy.isHit2(player)) continue; // プレイヤーと衝突していない
			float dx  = player.x - enemy.x;
			float dy  = player.y - enemy.y;
			float rot = Hell_atan2Ex(-dy, dx);
			if(!player.damage(rot, GH.PLAYER_DAMAGE_SPEED))
			{
				Hell_playWAV("damage");
				// プレイヤーダメージ
				damagePlayer();
				state = STATE_DAMAGE_WAIT;
				timer = GH.MAIN_DAMAGE_TIMER;
			}
		}
		// サウンドの再生
		if(bSndItem)   { Hell_stopWAV(5); Hell_playWAV("item", 0, 5);        } // アイテム取得音
		if(bSndEnemy)  { Hell_stopWAV(7); Hell_playWAV("destroy1", 0, 7);    } // 敵破壊音
		if(bSndEnemy2) { Hell_stopWAV(4); Hell_playWAV("destroy2", 0, 4);    } // 敵破壊音2
		if(bSndRefHit) { Hell_stopWAV(6); Hell_playWAV("reflect_hit", 0, 6); } // 打ち返しヒット

		// ボス自爆チェック
		if(Enemy.isBossSuicide())
		{
			timer = GH.MAIN_SUICIDE_TIMER;
			state = STATE_BOSS_SUICIDE;
			back.begin(Back.STATE_RASTER, GH.MAIN_SUICIDE_TIMER);
		}
	}
	void draw()
	{
		drawBack();
		message.draw();
		// 各トークンの描画
		switch(state)
		{
		case STATE_MENU:
			// メニューの描画
			drawMenu();
			return;
		case STATE_BOSS_SUICIDE:
			Enemy.drawBoss();
			break;
		default:
			Shot.parent.draw();
			Reflection.parent.draw();
			Enemy.parent.draw();
			player.draw(isDanger());
			Item.parent.draw();
			Bullet.parent.draw();
			break;
		}
		Particle.parent.draw();
		switch(state)
		{
		case STATE_MAIN:
		case STATE_DAMAGE_WAIT:
		case STATE_HIT_WAIT:
			drawOrbitForecast();
			break;
		default:
			break;
		}
		drawInfo();
		switch(state)
		{
		case STATE_GAMEOVER_INIT:
			Hell_setAlpha(HELL_ALPHA_ADD);
			Hell_drawRect(0, 0, GH.WIDTH, GH.HEIGHT, 0, 0, 0xff, 0, 0, 0xff*timer/TIMER_GAMEOVER_INIT);
			break;
		case STATE_GAMEOVER:
			Hell_drawFont("TIME OVER", 160, 200);
			break;
		case STATE_GAMECLEAR_INIT:
			Hell_setAlpha(HELL_ALPHA_ADD);
			int W2 = GH.WIDTH/2;
			int H2 = GH.HEIGHT/2;
			int alpha = 0x40*timer/TIMER_GAMECLEAR_INIT;
			Hell_drawRect(0, 0, GH.WIDTH, GH.HEIGHT, 0, 0, 0xff, 0xff, 0xff, alpha);
			break;
		case STATE_GAMECLEAR:
			int x = 160;
			int y = 200;
			if(timer > 0)
			{
				foreach(i; [0, 180])
				{
					int dx = cast(int)(200f*timer/TIMER_GAMECLEAR*Hell_cosEx(timer*4+i));
					int dy = cast(int)(200f*timer/TIMER_GAMECLEAR*-Hell_sinEx(timer*4+i));
					Hell_drawFont("GAME CLEAR", x+dx, y+dy, 1, 0xff, 0xff, 0xff, 0x80);
				}
			}
			else
			{
				Hell_drawFont("GAME CLEAR", x, y);
			}
			break;
		case STATE_SCORE_INIT:
			Hell_setAlpha(HELL_ALPHA_NORMAL);
			float ratio = 1 - 1.0 * timer / TIMER_SCORE_INIT;
			Hell_drawRect(0, 0, 640, 480, 0, 0, 0x00, 0x00, 0x00, cast(int)(0xa0*ratio));
			drawResult(ratio);
			break;
		case STATE_SCORE_MAIN:
			Hell_setAlpha(HELL_ALPHA_NORMAL);
			Hell_drawRect(0, 0, 640, 480, 0, 0, 0x00, 0x00, 0x00, 0xa0);
			drawResult(1);
			break;
		default:
			break;
		}
	}
	int clampAlpha(float alpha)
	{
		if(alpha > 0xff) return 0xff;
		return cast(int)alpha;
	}
	char[] getScoreRank(int score)
	{
		if(score < 3000)  return "D";
		if(score < 5000)  return "C";
		if(score < 8000)  return "B";
		if(score < 10000) return "A";
		else              return "S";
	}
	int getCountGradeRank(char[][] rankTbl)
	{
		int ret = 0;
		foreach(str; rankTbl)
		{
			switch(str)
			{
			case "S":
			case "A":
				ret++;
				break;
			default:
				break;
			}
		}
		return ret;
	}
	char[] getClearTimeString()
	{
		int min  = tPast/60/60;
		int sec  = (tPast/60)%60;
		int msec = (tPast%60)*1000/60;
		return format("%02d:%02d:%03d", min, sec, msec);
	}
	void drawResult(float ratio)
	{
//		Hell_drawTexture("result", 0, 0, 0, 0, 0, 0, 1f, 1f, 0, 0xff, 0xff, 0xff, clampAlpha(0xff*ratio));
		Hell_drawTexture("result", 0, 0);
		int x  = 350;
		int y  = 70;
		int dy = 44;
		int dx2 = 103;
		int dy2 = 8;
		int dx3 = 235;
		int dy3 = 0;
		float scale2 = 0.3f;
		
		bool[] blinkTbl = [
			scoreShot      >= 8000,
			scoreChain     >= 8000,
			scoreMissile   >= 8000,
			scoreMove      >= 8000,
			scoreLastTime  >= 8000,
			scoreClearTime >= 8000,
			scoreDamage    >= 8000,
		];
		bool[] scaleTbl = [
			scoreShot      >= 10000,
			scoreChain     >= 10000,
			scoreMissile   >= 10000,
			scoreMove      >= 10000,
			scoreLastTime  >= 10000,
			scoreClearTime >= 10000,
			scoreDamage    >= 10000,
		];
		char[][] rankTbl = [
			getScoreRank(scoreShot),
			getScoreRank(scoreChain),
			getScoreRank(scoreMissile),
			getScoreRank(scoreMove),
			getScoreRank(scoreLastTime),
			getScoreRank(scoreClearTime),
			getScoreRank(scoreDamage),
		];
		
		char[][] scoreTbl = [
			format("%6d", scoreShot),
			format("%6d", scoreChain),
			format("%6d", scoreMissile),
			format("%6d", scoreMove),
			format("%6d", scoreLastTime),
			format("%6d", scoreClearTime),
			format("%6d", scoreDamage),
		];
		char[][] scoreTbl2 = [
			format("%d", cntShot),
			format("%d", cntChain),
			format("%d", cntMissile),
			format("%d", cntMove),
			format("%02d:%02d:%03d", tMin, tSec, tMSec),
			getClearTimeString(),
			format("%d", cntDamage),
		];
		foreach(i, msg; scoreTbl)
		{
			int alpha = clampAlpha(0xff * ratio * 10);
			int b = 0xff;
			if(blinkTbl[i]) b = 0xa0 + cast(int)(0x60*Hell_sinEx(tCount*8));
			Hell_drawFontEx(x, y, .5f, 0xff, 0xff, b, alpha, msg);
			int c = clampAlpha(alpha - 30);
			drawFont2(scoreTbl2[i], x+dx2, y+32, c, c, c, alpha);
			Hell_drawFontEx(x+dx3, y+dy3, .8f, 0xff, 0xff, b, alpha, rankTbl[i]);
			if(ratio < 0.1) return;
			ratio -= 0.1f;
			y += dy;
		}
		int alpha = clampAlpha(0xff * ratio * 10);
		int x2 = 145;
		int y2 = 432;
		Hell_drawFontEx(x2, y2, 1f, 0xff, 0xff, 0xff, alpha, "%8d", scoreTotal);
		int x3 = 598;
		int y3 = 410;
		int cnt = getCountGradeRank(rankTbl);
		if(cnt > 0)
		{
			if(cnt > 4) cnt = 4;
			int d = 0;
			switch(cnt)
			{
			case 1: d = 2; break;
			case 2: d = 1; break;
			case 3: d = 1; break;
			case 4: d = 0; break;
			}
			float scale = 2.0f;
			scale += 2.0f * ratio;
			Hell_drawTextureEx("tutorial", x3, y3, 640+d*20, 0, 20, 20, scale, scale);
		}
	}
	// 背景の描画
	void drawBack()
	{
		back.draw();
		if(isDanger())
		{
			// 危険演出
			Hell_setAlpha(HELL_ALPHA_ADD);
			int H = 12;
			float r1 = Hell_sinEx((tLimit+90)%180);
			int alpha = cast(int)(0x80*r1);
			//Hell_drawRect(0, 0, GH.WIDTH, H, 0, 0, 0xff, 0, 0, alpha);
			//Hell_drawRect(0, GH.HEIGHT-H, GH.WIDTH, H, 0, 0, 0xff, 0, 0, alpha);
			Hell_drawRect(0, 160, GH.WIDTH, H, 0, 0, 0xff, 0, 0, alpha);
			Hell_drawRect(0, GH.HEIGHT-H-170, GH.WIDTH, H, 0, 0, 0xff, 0, 0, alpha);
			Hell_drawFontEx(0, 180, 2.23f, 0xff, 0, 0, 0x80, "%02d:%02d:%03d", tMin, tSec, tMSec);
		}
	}
	void drawMenu()
	{
		Hell_setAlpha(HELL_ALPHA_ADD);
		Hell_drawRect(0, MODE1_Y + selected*MODE_H+10, 640, MODE_H-20, 0, 0, 0x80, 0xff, 0x80, cast(int)(0x80*Hell_sinEx((timer*2)%180)));
		Hell_setAlpha(HELL_ALPHA_NORMAL);
		Hell_drawFont("RESTART",       MODE_X, MODE1_Y, .5f);
		Hell_drawFont("BACK TO TITLE", MODE_X, MODE2_Y, .5f);
		Hell_drawFont("CANCEL",        MODE_X, MODE3_Y, .5f);
		player.aim.draw();
	}
	// 起動予測の描画
	void drawOrbitForecast()
	{
		if(!player.aim.isShot()) return; // ショット撃てない

		foreach(e; Enemy.parent.getArray())
		{
			if(!e.exist)       continue;
			if(!e.isHitShot()) continue; // ショットが当たらない
			if(e.isBarrier())  continue; // バリアがある
			float dx  = e.x - player.aim.x;
			float dy  = e.y - player.aim.y;
			if(dx*dx + dy*dy > e.w*e.w + Shot.SIZE*Shot.SIZE) continue;
			
			float deg   = Hell_atan2(-dy, dx);
			float vx    = 16 * Hell_cos(deg);
			float vy    = 16 * -Hell_sin(deg);
			int   alpha = Hell_randint(0x80, 0xff);
			Hell_setAlpha(HELL_ALPHA_ADD);
			float px = e.x + vx;
			float py = e.y + vy;
			for(int i = 0; i < GH.SHOT_SEARCH_LENGTH; i++)
			{
				px += vx;
				py += vy;
				if(px < 0) { px = -px; vx = -vx; }
				if(py < 0) { py = -py; vy = -vy; }
				if(px > GH.WIDTH)  { px = GH.WIDTH  - (cast(int)px%GH.WIDTH);  vx = -vx; }
				if(py > GH.HEIGHT) { py = GH.HEIGHT - (cast(int)py%GH.HEIGHT); vy = -vy; }
				int s = Hell_randint(1, 8);
				Hell_drawRectEx(cast(int)px, cast(int)py, s, s, 0, 0, 0x00, 0xff, 0x00, alpha);
			}
		}
	}
	void drawInfo()
	{
		drawLimitTime();
		drawPastTime();
		
		if(player.isDamage())
		{
			Hell_drawTexture("tutorial", 640-48, 480-88, 640+32*((tPast%30)/15), 22, 32, 32, 1.5f, 1.5f);
		}
		if(!GH.DEBUG) return;
		//  デバッグ用
		Hell_setAlpha(HELL_ALPHA_NORMAL);
		Hell_drawFontEx(0, 40,  .4f, 0xff, 0xff, 0xff, 0xff, "SHOT     :%d", Shot.parent.count());
		Hell_drawFontEx(0, 60,  .4f, 0xff, 0xff, 0xff, 0xff, "REFLECT  :%d", Reflection.parent.count());
		Hell_drawFontEx(0, 80,  .4f, 0xff, 0xff, 0xff, 0xff, "ENEMY    :%d", Enemy.parent.count());
		Hell_drawFontEx(0, 100, .4f, 0xff, 0xff, 0xff, 0xff, "BULLET   :%d", Bullet.parent.count());
		Hell_drawFontEx(0, 120, .4f, 0xff, 0xff, 0xff, 0xff, "PARTICLE :%d", Particle.parent.count());
		Hell_drawFontEx(0, 140, .4f, 0xff, 0xff, 0xff, 0xff, "RATIO    :%d", Ratio);
		Hell_drawFontEx(0, 160, .4f, 0xff, 0xff, 0xff, 0xff, "MOVE  CNT:%d", cntMove);
		Hell_drawFontEx(0, 180, .4f, 0xff, 0xff, 0xff, 0xff, "SHOT  CNT:%d", cntShot);
		Hell_drawFontEx(0, 200, .4f, 0xff, 0xff, 0xff, 0xff, "CHAIN MAX:%d", cntChain);
		Hell_drawFontEx(0, 220, .4f, 0xff, 0xff, 0xff, 0xff, "MISSILE  :%d", cntMissile);
		Hell_drawFontEx(0, 240, .4f, 0xff, 0xff, 0xff, 0xff, "DAMAGECNT:%d", cntDamage);
	}
	void drawLimitTime()
	{
		int dy = -60 * tInit / TIMER_INIT;
		int g = 0xff;
		int b = 0xff;
		Hell_setAlpha(HELL_ALPHA_NORMAL);
		if(isDanger())
		{
			g = b = 8*(tLimit%32);
		}
		Hell_drawFontEx(0, dy, .4f, 0xff, g, b, 0xff, "LIMIT- %02d:%02d:%03d", tMin, tSec, tMSec);
		
		int OX = 210;
		int OY = 12;
		int W  = GH.WIDTH-OX-16;
		int a  = 0xa0;
		Hell_drawRect(OX, OY+dy, W, 3, 0, 0, 0xff, g, b, a);
		Hell_drawCircle(OX, OY+1+dy, 4, 0, 0xff, g, b, a);
		Hell_drawCircle(OX+W, OY+dy, 4, 0, 0xff, g, b, a);
		int ofs = player.isDamage() ? 2 : (tLimit%40)/20;
		Hell_drawTextureEx("player", OX + W*tLimit/tLimitMax, OY+dy, 128*ofs, 0, 128, 128, 0.2f, 0.2f);
	}
	void drawPastTime()
	{
		Hell_setAlpha(HELL_ALPHA_NORMAL);
		if(tRecordSmash > 0)
		{
			if(tRecordSmash < TIMER_RECORD_SMASH) Hell_setAlpha(HELL_ALPHA_ADD);
			int c = Hell_randint(0x00, tRecordSmash*4)%0xff;
			int a = 0xff*tRecordSmash/TIMER_RECORD_SMASH;
			Hell_drawFontEx(GH.WIDTH/2-210, GH.HEIGHT/2+24+12, 1f, 0xff, c, c, a, "NEW RECORD!!!");
		}

		int dx = 180 * tInit / TIMER_INIT;
		Hell_setAlpha(HELL_ALPHA_NORMAL);
		if(tRecordSmash%4 < 2)
		{
			Hell_drawFontEx(GH.WIDTH-170+dx, GH.HEIGHT-40, .4f, 0xff, 0xff, 0x00, 0xff, "NOW  %02d:%02d:%03d", tPast/60/60, (tPast/60)%60, 1000*(tPast%60)/60);
		}
		Hell_drawFontEx(GH.WIDTH-170+dx, GH.HEIGHT-20, .4f, 0xff, 0xff, 0x00, 0xff, "BEST %02d:%02d:%03d", tBest/60/60, (tBest/60)%60, 1000*(tBest%60)/60);
	}
}

/**
 * タイトルシーン
 */
class SceneTitle
{
public:
	enum
	{
		STATE_INIT,     // 初期化
		STATE_MAIN,     // メイン
		STATE_TUTORIAL, // チュートリアル
		STATE_STAFF,    // スタッフ
		STATE_BLINK,    // 点滅
	}
	static int   nSkin;            // スキン番号
	const int TIMER_CHARA  = 400;
	const int TIMER_BLINK  = 90;
	const int TIMER_TROPHY = 20;
	const int TIMER_SCORE  = 60;
	const int SKIN_MAX = 5;
	const int MODE_H  = 40;
	const int MODE_X  = 120;
	const int MODE0_Y = 200;
	const int MODE1_Y = 200 + MODE_H * 1;
	const int MODE2_Y = 200 + MODE_H * 2;
	const int MODE3_Y = 200 + MODE_H * 3;
	const int MODE4_Y = 200 + MODE_H * 4;
	const int MODE5_Y = 200 + MODE_H * 5;
	const char[] strLogo = "NATSUKI BOOST 2";
	int          timer;
	int          state;
	int          selected; // 選択中の項目
	int          tLogo;
	int          tChara;   // キャラカットインタイマ
	int          tTrophy;  // トロフィータイマ
	ConfigParser savedata;
	int[]        idxTrophy; // トロフィーインデックス
	bool         bSkinChange; // スキンチェンジできるかどうか
	Token[]      token;
	int          tScore;
public:
	this()
	{
		nSkin = 0;
		savedata = new ConfigParser();
		idxTrophy = new int[SKIN_MAX];
		token = new Token[4];
		foreach(inout t; token)
		{
			t = new Token();
		}
	}
	char[] getTimerString(int t)
	{
		int min  = t/60/60;
		int sec  = (t/60)%60;
		int msec = 1000*(t%60)/60;
		return format("%02d:%02d:%03d", min, sec, msec);
	}
	void init()
	{
		try
		{
			savedata.load(GH.SAVE_DATA);
		}
		catch(Exception e)
		{
			savedata.set("EASY",         59*60*60 + 59*60 + 59);
			savedata.set("NORMAL",       59*60*60 + 59*60 + 59);
			savedata.set("HARD",         59*60*60 + 59*60 + 59);
			savedata.set("HELL",         59*60*60 + 59*60 + 59);
			savedata.set("EASY_SCORE",   0);
			savedata.set("NORMAL_SCORE", 0);
			savedata.set("HARD_SCORE",   0);
			savedata.set("HELL_SCORE",   0);
			savedata.set("HELL_CLEAR",   false);
			savedata.set("EASY_GRADE",   0);
			savedata.set("NORMAL_GRADE", 0);
			savedata.set("HARD_GRADE",   0);
			savedata.set("HELL_GRADE",   0);
			savedata.write();
		}
		Hell_setBgmVolume(128);
		Hell_wait(100);
		Hell_playBgm("resource/sound/99.ogg");
		timer    = 0;
		state    = STATE_INIT;
		selected = 0;
		tLogo    = 0;
		tChara   = TIMER_CHARA;
		tTrophy  = TIMER_TROPHY;
		tScore   = 0;
		for(int i = 0; i < SKIN_MAX; i++)
		{
			int type  = -1;
			int t2    = 999999;
			switch(i)
			{
			case 0: t2 = 3; break;
			case 1:
				t2 = savedata.getint("EASY_GRADE");
				break;
			case 2:
				t2 = savedata.getint("NORMAL_GRADE");
				break;
			case 3:
				t2 = savedata.getint("HARD_GRADE");
				break;
			case 4:
				t2 = savedata.getint("HELL_GRADE");
				break;
			default:
				break;
			}
			if(t2 >= 4)      { type = 0; }
			else if(t2 >= 2) { type = 1; }
			else if(t2 >= 1) { type = 2; }
			idxTrophy[i] = type;
		}
		bSkinChange = false;
		foreach(i, idx; idxTrophy)
		{
			if(i == 0) continue;
			if(idx == 0) bSkinChange = true;
		}
		foreach(t; token)
		{
			t.create(Hell_randint(0, 640), Hell_randint(0, 480), 0, 0, 0, 0, 0, 0);
		}
	}
	void update()
	{
		if(tScore > 0) tScore--;
		{
			float x = Hell_getMouseX() + 64*Hell_cosEx(timer*5);
			float y = Hell_getMouseY() + 64*-Hell_sinEx(timer*5);
			foreach(t; token)
			{
				float dx = x - t.x;
				float dy = y - t.y;
				float rot = Hell_atan2Ex(-dy, dx);
				t.vx += Hell_cosEx(rot);
				t.vy += -Hell_sinEx(rot);
				t.vx *= 0.9f;
				t.vy *= 0.9f;
				t.rot += 2;
				t.move();
				x = t.x;
				y = t.y;
			}
		}
		switch(state)
		{
		case STATE_INIT:
		case STATE_MAIN:
			int t = tLogo/10;
			if(t < strLogo.length)
			{
				switch(t)
				{
				case 7, 13:
					break;
				case 14:
					if(tLogo%10 == 1) Hell_playWAV("kin");
					break;
				default:
					if(tLogo%10 == 1) Hell_playWAV("title");
					break;
				}
			}
			break;
		default:
			break;
		}
		
		switch(state)
		{
		case STATE_INIT:
			tChara = tChara * 90 / 100;
			tLogo++;
			timer++;
			if(isPushEnter())
			{
				playClick();
				state = STATE_MAIN;
			}
			break;
		case STATE_MAIN:
			if(tTrophy > 0) tTrophy--;
			tChara = tChara * 90 / 100;
			tLogo++;
			timer++;
			if(isPushEnter())
			{
				playClick();
				if(selected == -1)
				{
					state = STATE_TUTORIAL;
				}
				else if(selected == 4)
				{
					state = STATE_STAFF;
				}
				else
				{
					gMode = selected;
					state = STATE_BLINK;
					timer = TIMER_BLINK;
				}
			}
			if(isPushCancel())
			{
				if(bSkinChange)
				{
					while(true)
					{
						nSkin++;
						if(nSkin >= SKIN_MAX)
						{
							nSkin = 0;
							break;
						}
						if(idxTrophy[nSkin] == 0)
						{
							break;
						}
					}
					char[][] DIR_TBL = [
						"resource/image/",
						"resource/image1/",
						"resource/image2/",
						"resource/image3/",
						"resource/image4/",
					];
					Hell_playWAV("kin");
					loadImage(DIR_TBL[nSkin]);
				}
			}
			int my = Hell_getMouseY();
			bool bChange = false;
			if(MODE0_Y <= my && my < MODE0_Y + MODE_H) { bChange = selected != -1; selected = -1; }
			if(MODE1_Y <= my && my < MODE1_Y + MODE_H) { bChange = selected != 0;  selected = 0; }
			if(MODE2_Y <= my && my < MODE2_Y + MODE_H) { bChange = selected != 1;  selected = 1; }
			if(MODE3_Y <= my && my < MODE3_Y + MODE_H) { bChange = selected != 2;  selected = 2; }
			if(MODE4_Y <= my && my < MODE4_Y + MODE_H) { bChange = selected != 3;  selected = 3; }
			if(savedata.getboolean("HELL_CLEAR"))
			{
				if(MODE5_Y <= my && my < MODE5_Y + MODE_H) { bChange = selected != 4;  selected = 4; }
			}
			if(bChange) playClick();
			break;
		case STATE_TUTORIAL:
			{
				int mx = Hell_getMouseX();
				int my = Hell_getMouseY();
				int px = 85;
				int py = 192;
				if(px - 5 <= mx && mx <= px + 5 && py - 5 <= my && my <= py + 5)
				{
					if(tScore < 1) tScore = TIMER_SCORE;
				}
			}
		case STATE_STAFF:
			tLogo++;
			timer++;
			if(isPushEnter())
			{
				playClick();
				state = STATE_MAIN;
			}
			break;
		case STATE_BLINK:
			tChara += timer/4;
			timer--;
			if(isPushEnter()) timer = 0;
			if(timer < 1)
			{
				gotoMain();
			}
			break;
		}
		gSceneMain.player.aim.update();
	}
	void playClick()
	{
		Hell_stopWAV(7);
		Hell_playWAV("click", 0, 7);
	}
	void draw()
	{
		Hell_setAlpha(HELL_ALPHA_NORMAL);
		// 背景
		float scale = 1.f + Hell_abs(.03f*Hell_sinEx(timer));
		Hell_drawTextureEx("back", 320, 240, 0, 0, 640, 480, scale, scale, Hell_sinEx(timer), 0xff, 0xff, 0xff, 0xc0);
		// キャラ
		Hell_drawTexture("title", GH.WIDTH-400+tChara, 0);
		// ロゴ
		Hell_setAlpha(HELL_ALPHA_NORMAL);
		int t = tLogo/10;
		if(t > strLogo.length) t = strLogo.length;
		if(t == strLogo.length)
		{
			int d = 4;
			Hell_drawFontEx(80+d, 40+d, 1, 0xff, 0xff, 0x80, 0x80, strLogo);
		}
		Hell_drawFontEx(80, 40, 1, 0xff, 0xff, 0xff, 0xff, strLogo[0..t]);
		if(t < strLogo.length)
		{
			float s = 1f + 0.1f*(10-tLogo%10);
			int   a = 0xff * (tLogo%10) / 10;
			Hell_drawFontEx(80+t*32, 40, s, 0xff, 0xff, 0xff, a, strLogo[t]);
		}
		// メニュー
		switch(state)
		{
		case STATE_INIT:
			if(timer%128 < 100)
			{
				Hell_drawFont("PRESS LEFT MOUSE BUTTON", 140, 360, .5f, timer%0xff, timer/2%0xff, timer/3%0xff);
			}
			break;
		case STATE_MAIN:
			Hell_setAlpha(HELL_ALPHA_ADD);
			Hell_drawRect(0, MODE1_Y + selected*MODE_H+10, 640, MODE_H-20, 0, 0, 0x80, 0xff, 0x80, cast(int)(0x80*Hell_sinEx((timer*2)%180)));
			Hell_setAlpha(HELL_ALPHA_NORMAL);
			Hell_drawTextureEx("player", MODE_X-32, MODE1_Y + 16 + selected*MODE_H, 128*((timer%60)/30), 0, 128, 128, 0.4f, 0.4f);
			int cntItem = savedata.getboolean("HELL_CLEAR") ? 5 : 4;
			for(int i = -1; i < cntItem; i++)
			{
				drawMenu(i);
				int type = -1;
				switch(i)
				{
				case 0, 1, 2, 3, 4:
					type = idxTrophy[i];
					break;
				default:
					break;
				}
				if(type >= 0) drawTrophy(i, type);
			}
			drawSkinInfo();
			break;
		case STATE_TUTORIAL:
			Hell_setAlpha(HELL_ALPHA_NORMAL);
			Hell_drawTexture("tutorial", 0, 0);
			break;
		case STATE_STAFF:
			Hell_setAlpha(HELL_ALPHA_NORMAL);
			Hell_drawTexture("staff", 0, 0);
			break;
		case STATE_BLINK:
			if(timer < 45 || timer%8 < 4) drawMenu(selected);
			break;
		}
		foreach(i, t2; token)
		{
			if(idxTrophy[i+1] != 0) continue;
			int ox, oy, ow, oh;
			switch(i)
			{
			case 0:  ox = 0;   oy = 0;   ow = 128; oh = 128; break; // なす
			case 1:  ox = 128; oy = 0;   ow = 128; oh = 128; break; // 5
			case 2:  ox = 0;   oy = 128; ow = 128; oh = 128; break; // だいこん
			case 3:  ox = 128; oy = 128; ow = 128; oh = 128; break; // たこやき
			}
			Hell_drawTextureEx("enemy", t2.x, t2.y, ox, oy, ow, oh, .2f, .2f, t2.rot);
		}
		gSceneMain.player.aim.draw();
		if(tScore > 0)
		{
			for(int i = 0; i < 8; i++)
			{
				float ratio = Hell_abs(1f - 1f * (tScore + TIMER_SCORE/2) / TIMER_SCORE);
				float px = 85  - 8;
				float py = 192 - 8;
				float dx = ratio * 100 * Hell_cosEx(i*45+tScore*4);
				float dy = ratio * 100 * -Hell_sinEx(i*45+tScore*4);
				Hell_drawFont("10", cast(int)(px + dx), cast(int)(py + dy), .5f, 0xff, 0xff, 0xff, 0xaf+0x40*tScore/TIMER_SCORE);
			}
		}
	}
	void drawSkinInfo()
	{
		if(bSkinChange)
		{
			Hell_setAlpha(HELL_ALPHA_ADD);
			Hell_drawRect(0, 480-24, 640, 24, 0, 0, 0, 0x80, 0);
			Hell_setAlpha(HELL_ALPHA_NORMAL);
			Hell_drawTexture("tutorial", 640-48, 480-48, 640+32*((timer%60)/30), 22, 32, 32, 1.5f, 1.5f);
			char[] name = "";
			switch(nSkin)
			{
			case 0:
				name = "NATSUKI";
				break;
			case 1:
				name = "?????";
				break;
			case 2:
				name = "G.S.";
				break;
			case 3:
				name = "F.A.";
				break;
			case 4:
				name = "ASTRO";
				break;
			}
			Hell_drawFont(name, 640 - 140, 480 - 20, .4f);
		}
	}
	void drawTrophy(int i, int type)
	{
		int dx = -20 + 11;
		int dy = 2   + 11;
		float scale = 1.0f + 1.0f * tTrophy / TIMER_TROPHY;
		switch(i)
		{
		case 0:
			break;
		case 1:
			Hell_drawTextureEx("tutorial", MODE_X+dx, MODE1_Y+dy, 640+type*20, 0, 20, 20, scale, scale);
			break;
		case 2:
			Hell_drawTextureEx("tutorial", MODE_X+dx, MODE2_Y+dy, 640+type*20, 0, 20, 20, scale, scale);
			break;
		case 3:
			Hell_drawTextureEx("tutorial", MODE_X+dx, MODE3_Y+dy, 640+type*20, 0, 20, 20, scale, scale);
			break;
		case 4:
			Hell_drawTextureEx("tutorial", MODE_X+dx, MODE4_Y+dy, 640+type*20, 0, 20, 20, scale, scale);
			break;
		}
	}
	void drawMenu(int i)
	{
		switch(i)
		{
		case -1:
			Hell_drawFontEx(MODE_X, MODE0_Y, .5f, 0xff, 0xff, 0xff, 0xff, "TUTORIAL");
			break;
		case 0:
			Hell_drawFontEx(MODE_X, MODE1_Y, .5f, 0xff, 0xff, 0xff, 0xff, "EASY   %s %09d", getTimerString(savedata.getint("EASY")), savedata.getint("EASY_SCORE"));
			break;
		case 1:
			Hell_drawFontEx(MODE_X, MODE2_Y, .5f, 0xff, 0xff, 0xff, 0xff, "NORMAL %s %09d", getTimerString(savedata.getint("NORMAL")), savedata.getint("NORMAL_SCORE"));
			break;
		case 2:
			Hell_drawFontEx(MODE_X, MODE3_Y, .5f, 0xff, 0xff, 0xff, 0xff, "HARD   %s %09d", getTimerString(savedata.getint("HARD")), savedata.getint("HARD_SCORE"));
			break;
		case 3:
			Hell_drawFontEx(MODE_X, MODE4_Y, .5f, 0xff, 0xff, 0xff, 0xff, "HELL   %s %09d", getTimerString(savedata.getint("HELL")), savedata.getint("HELL_SCORE"));
			break;
		case 4:
			Hell_drawFontEx(MODE_X, MODE5_Y, .5f, 0xff, 0xff, 0xff, 0xff, "STAFF");
			break;
		}
	}
}

enum
{
	STATE_TITLE,
	STATE_MAIN,
}
int        gMode;       // ゲームモード
int        gState;      // シーン状態
SceneTitle gSceneTitle; // タイトルシーン
SceneMain  gSceneMain;  // メインシーン
// メインシーンへ遷移
void gotoMain()
{
	gState = STATE_MAIN;
	gSceneMain.init();
}
// タイトルシーンへ遷移
void gotoTitle()
{
	gState = STATE_TITLE;
	gSceneTitle.init();
}

void loadImage(char[] dir)
{
	Hell_loadTexture("back",     dir ~ "back.bmp",     [0,  0xff, 0]);
	Hell_loadTexture("player",   dir ~ "player.bmp",   [-2, 0, 0]);
	Hell_loadTexture("enemy",    dir ~ "enemy.bmp",    [-2, 0, 0]);
	Hell_loadTexture("enemy2",   dir ~ "enemy2.bmp",   [-2, 0, 0]);
}
// ゲーム初期化
void init()
{
	// リソース読み込み
	loadImage("resource/image/");
	Hell_loadTexture("title",    "resource/image/title.bmp",   [0,  0xff, 0]);
	Hell_loadTexture("tutorial", "resource/image/tutorial.bmp", [0xff, 0, 0xff]);
	Hell_loadTexture("tama",     "resource/image/tama.bmp",     [0,  0xff, 0]);
	Hell_loadTexture("shot",     "resource/image/shot.bmp",     [-1, 0, 0]);
	Hell_loadTexture("effect",   "resource/image/effect.bmp",   [-1, 0, 0]);
	Hell_loadTexture("result",   "resource/image/result.bmp",   [0, 0xff, 0]);
	Hell_loadTexture("staff",    "resource/image/staff.bmp",    [0, 0xff, 0]);
	Hell_loadTexture("font2",    "resource/image/font2.bmp",    [-2, 0, 0]);
	Hell_loadWAV("shot",         "resource/sound/shot.wav");
	Hell_loadWAV("destroy1",     "resource/sound/destroy1.wav");
	Hell_loadWAV("destroy2",     "resource/sound/destroy2.wav");
	Hell_loadWAV("destroy3",     "resource/sound/destroy3.ogg");
	Hell_loadWAV("move",         "resource/sound/move.wav");
	Hell_loadWAV("item",         "resource/sound/item.wav");
	Hell_loadWAV("click",        "resource/sound/click.wav");
	Hell_loadWAV("crash",        "resource/sound/crash.wav");
	Hell_loadWAV("damage",       "resource/sound/damage.wav");
	Hell_loadWAV("clock",        "resource/sound/clock.wav");
	Hell_loadWAV("reflect_hit",  "resource/sound/reflect_hit.wav");
	Hell_loadWAV("pi",           "resource/sound/pi.wav");
	Hell_loadWAV("title",        "resource/sound/title.wav");
	Hell_loadWAV("clear",        "resource/sound/clear.wav");
	Hell_loadWAV("kin",          "resource/sound/kin.wav");
	for(int i = 1; i < 11; i++)
	{
		Hell_loadWAV(format("v%d", i), format("resource/sound/v%d.wav", i));
	}

	gSceneTitle = new SceneTitle();
	gSceneMain  = new SceneMain();
	gSceneTitle.init();
}

// ゲームループ
void loop()
{
	switch(gState)
	{
	case STATE_TITLE:
		gSceneTitle.update();
		gSceneTitle.draw();
		break;
	case STATE_MAIN:
		gSceneMain.update();
		gSceneMain.draw();
		break;
	}
}


// ユーティリティ
void drawFont2(char[] msg, int x, int y, int r=0xff, int g=0xff, int b=0xff, int a=0xff)
{
	const int WIDTH  = 12; // フォントの幅
	const int HEIGHT = 15; // フォントの高さ
	char decoder[]   = "0123456789:";
	foreach(c; msg)
	{
		int idx = find(decoder, c); // 文字のインデックスを取得
		if(c == ' ' | idx < 0)
		{
			x += WIDTH; // スペース
			continue;
		}
		Hell_drawTextureEx("font2", x, y, idx*WIDTH, 0, WIDTH, HEIGHT, 1f, 1f, 0, r, g, b, a);
		x += WIDTH;
	}
}
bool isPressEnter()
{
	if(Hell_isPressKey(HELL_z)) return true;
	if(Hell_isPressJButton(0))  return true;
	if(Hell_isPressMouse() & HELL_BUTTON_LEFT) return true;
	return false;
}
bool isPushEnter()
{
	if(Hell_isPushKey(HELL_z)) return true;
	if(Hell_isPushJButton(0))  return true;
	if(Hell_isPushMouse() & HELL_BUTTON_LEFT) return true;
	return false;
}
bool isPressCancel()
{
	if(Hell_isPressKey(HELL_x)) return true;
	if(Hell_isPressJButton(1))  return true;
	if(Hell_isPressMouse() & HELL_BUTTON_RIGHT) return true;
	return false;
}
bool isPushCancel()
{
	if(Hell_isPushKey(HELL_x)) return true;
	if(Hell_isPushJButton(1))  return true;
	if(Hell_isPushMouse() & HELL_BUTTON_RIGHT) return true;
	return false;
}
bool isPushMenu()
{
	if(Hell_isPushKey(HELL_c)) return true;
	if(Hell_isPushJButton(2))  return true;
	if(Hell_isPushMouse() & HELL_BUTTON_MIDDLE) return true;
	return false;
}

int Hell_main(char[][] args)
{
	// 設定ファイル読み込み
	GH.load();
	try
	{
		// 画面を作る
		if(args.length > 1)
		{
			Hell_init("NATSUKI BOOST 2", GH.WIDTH, GH.HEIGHT, true);
		}
		else
		{
			Hell_init("NATSUKI BOOST 2", GH.WIDTH, GH.HEIGHT);
		}
		init();
		while(true)
		{
			Hell_begin();
			loop();
			if(GH.DEBUG) Hell_drawFPS(280, GH.HEIGHT-32);
			Hell_update();
			Hell_wait(g_fps);
		}
	}
	catch(Exception e)
	{
		Hell_write(e);
	}
	finally
	{
		Hell_quit();
	}
	
	return 0;
}


public int main(char[][] args)
{
	return Hell_main(args);
}

